/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef EPC_EVENT_H
#define EPC_EVENT_H

#include <string.h>
#include <systemd/sd-id128.h>
#include <time.h>

#include "uref.h"
#include "list.h"
#include "notify_queue.h"
#include "module.h"

#define EV_BOOT_EVENT "bev"
#define EV_TIMESTAMP "ts"
#define EV_TYPE_NAME "tn"

#define BSON_EVENT_ID "bson_event"

struct epc_event;

struct epc_event_ops {
	void (*release)(struct epc_event *);
	void (*serialize)(struct epc_event*, struct epc_object*);
};

struct epc_event_type {
	char *name;

	struct epc_event_ops default_ops;

	/* Add function to allocate event from DB */

	/* Allocate event based on passed data */
	int (*allocate_event)(struct epc_event_type *type,
						  void *data, struct epc_event **ev);

	int (*deserialize_event)(struct epc_event_type *type,
							 struct epc_object *data, struct epc_event **ev);

	/* To be used by event factory */
	struct list_head node;
};

struct eq_node {
	struct epc_event *ev;
	struct nqueue_node node;
};

struct epc_event {
	struct timespec timestamp;
	struct system_booted_event *boot_event;

	/* TODO: add here some id field */

	struct epc_event_type *type;
	struct epc_event_ops ops;

	struct uref uref;
	/* To be used by event holder */
	struct list_head node;
	/* To be used by event processor */
	struct eq_node event_queue_node;

	epc_oid_t oid;
};

int epc_event_type_register(struct epc_event_type *type);
void epc_event_type_unregister(struct epc_event_type *type);

#define EPC_EVENT_TYPE_REGISTER(EVENT_TYPE, NAME)				\
static int CONCAT(CONCAT(event_type_, EPC_MODNAME_T), _init)(struct epc_module *module, \
									  struct epc_config *config, \
									  sd_event *event_loop)			\
{																	\
	return epc_event_type_register(&EVENT_TYPE);                 \
}																	\
																	\
static void CONCAT(CONCAT(event_type_, EPC_MODNAME_T), _cleanup)(struct epc_module *module) \
{																	\
	epc_event_type_unregister(&EVENT_TYPE);                      \
}																	\
																	\
static struct epc_module CONCAT(CONCAT(event_type_, EPC_MODNAME_T), _module) = {		\
	.name = EPC_MODNAME,											\
	.type = EPC_MODULE_TYPE_EVENT,                               \
	.init = CONCAT(CONCAT(event_type_, EPC_MODNAME_T), _init),	\
	.cleanup = CONCAT(CONCAT(event_type_, EPC_MODNAME_T), _cleanup), \
	.node = LIST_HEAD_INIT(CONCAT(CONCAT(event_type_, EPC_MODNAME_T), _module.node)), \
};																	\
																	\
EPC_MODULE_REGISTER(&CONCAT(CONCAT(event_type_, EPC_MODNAME_T), _module))

int epc_event_create(const char *type, void *data, struct epc_event **ev);

static inline void epc_event_ref(struct epc_event *ev)
{
	uref_get(&ev->uref);
}

static inline void epc_event_unref(struct epc_event *ev)
{
	uref_put(&ev->uref);
}

static inline int epc_event_is_of_type(struct epc_event *ev,
											  const char *type)
{
	return strcmp(ev->type->name, type) == 0;
}

char *epc_event_to_string(struct epc_event *ev);

static inline void epc_event_serialize(struct epc_event *ev, struct epc_object *out)
{
	ev->ops.serialize(ev, out);
}

int epc_event_init_internal(struct epc_event_type *ev_type,
							   struct epc_event *ev);

static inline void epc_event_cleanup_internal(struct epc_event *ev) {}

void epc_event_serialize_internal(struct epc_event *ev, struct epc_object *out);

int epc_event_deserialize_internal(const struct epc_object *in, struct epc_event_type *ev_type, struct epc_event *ev);

void set_boot_event(struct system_booted_event *sb_ev);
struct system_booted_event *get_boot_event(void);

int append_epc_event(struct nqueue_head *head, struct epc_event *ev);
struct epc_event *pop_epc_event(struct nqueue_head *head);

#endif /* EPC_EVENT_H */
