/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "action.h"
#include "action_executor.h"
#include "decision_made_event.h"
#include "unit_control_event.h"
#include "log.h"
#include "systemd_dbus.h"
#include "common.h"

#include <libsyscommon/dbus-system.h>

struct unit_action_data {
	struct action_executed_event *exe_info;
	bool wait_for_unit;
	char *job;
};

static int finish_action(struct unit_action_data *data, int result, const char *err_msg)
{
	struct decision_made_event *dm_ev = to_decision_made_event(data->exe_info->reason);
	int ret;

	if (err_msg)
		epc_object_append_string(data->exe_info->action_log, "error", err_msg);

	if (dm_ev->on_finished)
		dm_ev->on_finished(dm_ev, result);

	data->exe_info->result = result;
	ret = event_processor_report_event(&data->exe_info->event);
	epc_event_unref(&data->exe_info->event);
	free(data);

	return ret;
}

static void on_job_removed(GDBusConnection *connection,
		const gchar *sender_name,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *signal_name,
		GVariant *parameters,
		gpointer userdata)
{
	struct unit_action_data *data = userdata;
	char *job;

	assert(data);

	g_variant_get(parameters, "(uoss)", NULL, &job, NULL, NULL);

	if (strcmp(data->job, job) == 0)
		 finish_action(data, 0, NULL);
}

static void unit_action_handler(GVariant *var, void *userdata, GError *err)
{
	struct unit_action_data *data = userdata;
	dbus_handle_h bus;
	int ret;

	if (err) {
		finish_action(data, -err->code, err->message);
		return;
	}

	if (!data)
		return;

	bus = dbus_handle_get_connection(G_BUS_TYPE_SYSTEM, 0);
	if (!bus) {
		finish_action(data, -EINVAL, "Could not acquire systemd bus");
		return;
	}

	if (data->wait_for_unit) {
		ret = subscribe_dbus_signal(bus,
				"sender='org.freedesktop.systemd1",
				"org.freedesktop.systemd1.Manager",
				"JobRemoved",
				on_job_removed,
				data,
				NULL);
		if (ret < 0) {
			finish_action(data, ret, "Could not register for JobRemoved signal");
			return;
		}

		return;
	}

	finish_action(data, 0, NULL);
	return;
}

static int start_unit(struct epc_action *action,
		struct action_executed_event *exe_info)
{
	struct epc_event *ev = pop_epc_event(&action->execution_queue);
	struct decision_made_event *dm_ev = to_decision_made_event(ev);
	char *unit_name = NULL;
	char *unit_action = NULL;
	int ret;
	struct unit_action_data *data;

	/*
	 * We are passing an event to exe_info, so there is no need to unref it
	 * in this function
	 */
	exe_info->reason = ev;

	ret = epc_object_get_string(dm_ev->action_data,
			EPC_AD_UNIT_NAME, &unit_name);
	if (!unit_name) {
		epc_object_append_string(exe_info->action_log, "error",
				"Unit name not specified");
		exe_info->result = -EINVAL;
		return 0;
	}

	data = malloc(sizeof(*data));
	if (!data)
		return -ENOMEM;

	data->exe_info = exe_info;

	ret = epc_object_get_bool(dm_ev->action_data, "wait", &data->wait_for_unit);
	if (ret < 0)
		data->wait_for_unit = false;

	ret = epc_object_get_string(dm_ev->action_data,
			EPC_AD_UNIT_ACTION, &unit_action);
	if (!unit_action)
		return finish_action(data, -EINVAL, "Unit action not specified");

	if (strcmp(unit_action, "StartUnit") &&
			strcmp(unit_action, "StopUnit") &&
			strcmp(unit_action, "RestartUnit"))
		return finish_action(data, -EINVAL,
				"Unit action is not one of \"StartUnit\", "
				"\"StopUnit\" or \"RestartUnit\"");

	ret = epc_dbus_call_systemd_simple_async(SYSTEMD_OBJ,
			SYSTEMD_MANAGER_INTERFACE,
			unit_action,
			unit_action_handler,
			data,
			"(ss)",
			unit_name,
			"replace");
	if (ret < 0)
		return finish_action(data, ret, "Failed to call systemd");
	else
		log_kmsg("Starting unit: %s", unit_name);

	/* The result will be delayed */
	exe_info->result = -EPROBE_DEFER;
	epc_event_ref(&exe_info->event);

	return 0;
}

static struct epc_action unit_start_action = {
	.action_id = EPC_ACTION_UNIT_START_ID,
	.impl_name = EPC_DEFAULT_ACTION_IMPL,
	.execute = start_unit,
	.node = LIST_HEAD_INIT(unit_start_action.node),
};

EPC_ACTION_REGISTER_SIMPLE(unit_start_action);
