/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <assert.h>
#include <errno.h>
#include <systemd/sd-id128.h>

#ifdef USE_EJDB
#include <ejdb/ejdb.h>
#endif

#include "database.h"
#include "log.h"
#include "module.h"

#define DEFAULT_DBADAPTER "nop_dbadapter"

static struct epc_database_adapter *database_adapter = NULL;

struct epc_database_adapter *epc_get_database_adapter(void)
{
	return database_adapter;
}

int epc_set_database_adapter(struct epc_database_adapter *adapter)
{
	if (database_adapter && adapter
		&& strcmp(DEFAULT_DBADAPTER, database_adapter->module.name) != 0)
		return -EINVAL;

	database_adapter = adapter;
	return 0;
}

int database_store(struct epc_object *obj, epc_oid_t *oid)
{
	if (!database_adapter)
		return -ENOSYS;

	return database_adapter->store(database_adapter, obj, oid);
}

int database_load(struct epc_object *query, struct epc_object *hints,
				  struct epc_object *result, uint32_t *nr)
{
	if (!database_adapter)
		return -ENOSYS;

	return database_adapter->load(database_adapter, query, hints, result, nr);
}

int database_get_well_known_oid(const char *name, epc_oid_t *oid)
{
	if (!database_adapter)
		return -ENOSYS;

	return database_adapter->get_well_known_oid(name, oid);
}

int database_get_by_oid(epc_oid_t *oid, struct epc_object *result)
{
	if (!database_adapter)
		return -ENOSYS;

	return database_adapter->get_by_oid(database_adapter, oid, result);
}

bool database_is_oid_valid(epc_oid_t *oid)
{
	if (!database_adapter)
		return false;

	return database_adapter->is_oid_valid(oid);
}

const char *database_get_id_key(void)
{
	if (!database_adapter)
		return NULL;

	return database_adapter->id_key;
}

int database_get_event(epc_oid_t *oid, struct epc_event **ev)
{
	struct epc_object *obj;
	int ret;

	if (database_adapter && database_adapter->get_event) {
		ret = database_adapter->get_event(oid, ev);
		if (ret == 0)
			(*ev)->oid = *oid;

		return ret;
	}

	ret = epc_object_new(&obj);
	if (ret) {
		log_error("Could not create object");
		return ret;
	}

	ret = database_get_by_oid(oid, obj);
	if (ret) {
		log_error("Could not find object");
		goto cleanup;
	}

	ret = epc_event_create(BSON_EVENT_ID, obj, ev);
	if (ret) {
		log_error("Could not create an event");
		goto cleanup;
	}

	(*ev)->oid = *oid;
	return 0;

cleanup:
	epc_object_unref(obj);
	return ret;
}

int database_dump(struct epc_object **out)
{
	struct epc_object *query, *query_child, *result;
	uint32_t num;
	int ret;

	ret = epc_object_new_flags(&query_child, EPC_FLAG_QUERY_MODE);
	if (ret < 0) {
		log_error("Unable to create epc_object");
		goto error;
	}

	ret = epc_object_append_string(query_child, "$not", "magicstring");
	if (ret < 0) {
		log_error("Unable to append string to epc_object");
		goto unref_query_child;
	}

	ret = epc_object_new_flags(&query, EPC_FLAG_QUERY_MODE);
	if (ret < 0) {
		log_error("Unable to create epc_object");
		goto unref_query_child;
	}

	ret = epc_object_append_object(query, database_get_id_key(), query_child);
	if (ret < 0) {
		log_error("Unable to append string to epc_object");
		epc_object_unref(query);
		goto unref_query_child;
	}

	epc_object_unref(query_child);

	ret = epc_object_new(&result);
	if (ret < 0) {
		log_error("Unable to create epc_object");
		epc_object_unref(query);
		goto error;
	}

	ret = database_load(query, NULL, result, &num);
	epc_object_unref(query);
	if (ret < 0) {
		log_error("Unable to load from database");
		epc_object_unref(result);
		goto error;
	}

	*out = result;
	return 0;

unref_query_child:
	epc_object_unref(query_child);
error:
	*out = NULL;
	return ret;
}
