/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <malloc.h>
#include <errno.h>

#include "decision_made_event.h"
#include "action.h"
#include "event_processor.h"
#include "log.h"
#include "service.h"
#include "common.h"
#include "database.h"
#include "json-config.h"

#include "vconf_key_changed_event.h"
#include "dbus_signal_event.h"

#define MODULE_NAME "activation_decision_maker"

struct activation_decision_maker {
	struct epc_event_handler eh;
	struct list_head rules;
};

enum rule_type {
	RULE_TYPE_VCONF,
	RULE_TYPE_DBUS
};

struct rule {
	const char *event;
	enum rule_type type;
	void *ev_data;
	bool match_oldval;
	bool match_newval;
	const char *action;
	const char *target;

	struct json_object *match;
	struct list_head node;
};

static const char *vkc_type_to_string(enum vkc_type t)
{
	switch (t) {
	case VKC_BOOL:
		return "bool";
	case VKC_INT:
		return "int";
	case VKC_DOUBLE:
		return "double";
	case VKC_STRING:
		return "string";
	default:
		return "unknown";
	}
}

static int activation_event_match(struct epc_event_handler *handler,
		struct epc_event *ev)
{
	if (strcmp(ev->type->name, VCONF_KEY_CHANGED_EVENT_ID) == 0 ||
			strcmp (ev->type->name, DBUS_SIGNAL_EVENT_ID) == 0)
			return 1;

	return 0;
}

static int vconf_field_cmp(struct vkc_value *a, struct vkc_value *b)
{
	if (a->type != b->type)
		return -1;

	switch (a->type){
	case VKC_BOOL:
		return a->value.b == b->value.b;
	case VKC_INT:
		return a->value.i == b->value.i;
	case VKC_DOUBLE:
		return a->value.d == b->value.d;
	case VKC_STRING:
		return strcmp(a->value.s, b->value.s) == 0;
	}
	return 0;
}

static int vconf_rule_match(struct rule *r, struct epc_event *event)
{
	struct vconf_key_changed_event *ev = to_vconf_key_changed_event(event);
	struct vconf_key_changed_event_data *ev_data = r->ev_data;
	int ret;

	if (ev_data->key_name && strcmp(ev_data->key_name, ev->key_name))
		return 0;

	if (r->match_oldval) {
		ret = vconf_field_cmp(&ev_data->oldval, &ev->oldval);
		if (ret < 0) {
			log_error("vconf type mismatch - oldval for key %s: %s instead of %s",
					ev_data->key_name,
					vkc_type_to_string(ev->oldval.type),
					vkc_type_to_string(ev_data->oldval.type));
		}

		if (ret <= 0)
			return 0;
	}

	if (r->match_newval) {
		ret = vconf_field_cmp(&ev_data->newval, &ev->newval);
		if (ret < 0) {
			log_error("vconf type mismatch - newval for key %s: %s instead of %s",
					ev_data->key_name,
					vkc_type_to_string(ev->newval.type),
					vkc_type_to_string(ev_data->newval.type));
		}

		if (ret <= 0)
			return 0;
	}

	return 1;
}

static int dbus_rule_match(struct rule *r, struct epc_event *event)
{
	struct dbus_signal_event *ev = to_dbus_signal_event(event);
	struct dbus_signal_event_data *ev_data = r->ev_data;

	if (ev_data->event_id && strcmp(ev_data->event_id, ev->event_id))
		return 0;

	if (ev->match_params && r->match && !epc_object_match_pattern(ev->match_params, r->match))
		return 0;

	return 1;
}

static int execute_rule_uc(struct rule *r, struct epc_event *ev, struct dm_event_data *ev_data)
{
	ev_data->action = EPC_ACTION_UNIT_START_ID;

	if (epc_fill_for_unit_start(ev_data->action_data, r->target)) {
		log_error("Unable to create event data");
		return -1;
	}
	epc_fill_for_unit_action(ev_data->action_data, r->action);

	return 0;
}

static int execute_rule_signal(struct rule *r, struct epc_event *ev, struct dm_event_data *ev_data)
{
	struct dbus_signal_event *dev = to_dbus_signal_event(ev);

	ev_data->action = EPC_ACTION_DBUS_SIGNAL_FORWARD_ID;

	if (epc_fill_for_signal_forward(ev_data->action_data, r->target,
			dev->path, dev->interface, dev->member)) {
		log_error("Unable to create event data");
		return -1;
	}

	return 0;
}

static int execute_rule(struct rule *r, struct epc_event *ev)
{
	struct dm_event_data ev_data = {
		.reason = ev,
		.who_made = MODULE_NAME,
	};
	struct epc_event *new_ev;
	int ret;

	ret = epc_object_new(&ev_data.action_data);
	if (ret < 0) {
		log_error("Could not create data object for action");
		goto out;
	}

	bool is_uc = !strcmp(r->action, "StartUnit") ||
			!strcmp(r->action, "StopUnit") ||
			!strcmp(r->action, "RestartUnit");
	bool is_sig = !strcmp(r->action, "ForwardDBusSignal");

	ret = is_uc ? execute_rule_uc(r, ev, &ev_data) :
		(is_sig ? execute_rule_signal(r, ev, &ev_data) : -1);
	if (ret < 0) {
		log_debug("Unsupported action");
		goto out;
	}

	ret = epc_event_create(DECISION_MADE_EVENT_ID, &ev_data, &new_ev);
	if (ret) {
		log_error("Unable to create event");
		goto out;
	}

	ret = event_processor_report_event(new_ev);
	epc_event_unref(new_ev);
	if (ret) {
		log_error("Unable to report event");
		goto out;
	}

out:
	epc_object_unref(ev_data.action_data);
	return ret;
}

static int activation_make_decision(struct epc_event_handler *handler)
{
	struct epc_event *ev = pop_epc_event(&handler->event_queue);
	struct activation_decision_maker *dm = container_of(handler,
			struct activation_decision_maker, eh);
	int ret;
	struct rule *r;

	list_for_each_entry(r, &dm->rules, node) {
		if (strcmp(r->event, ev->type->name))
			continue;

		if ((!strcmp(ev->type->name, VCONF_KEY_CHANGED_EVENT_ID) == 0 ||
				(r->type == RULE_TYPE_VCONF && !vconf_rule_match(r, ev))) &&
			(!strcmp(ev->type->name, DBUS_SIGNAL_EVENT_ID) == 0 ||
				(r->type == RULE_TYPE_DBUS && !dbus_rule_match(r, ev))))
			continue;

		ret = execute_rule(r, ev);
		if (ret < 0)
			break;
	}

	epc_event_unref(ev);
	return 0;
}

static int parse_vconf_field(json_object *root, const char *key, struct vkc_value *value)
{
	struct json_object *node = NULL;
	enum json_type type;
	const char *str;

	if (!json_object_object_get_ex(root, key, &node)) {
		log_debug("Config does not contain %s parameter", key);
		return -1;
	}

	type = json_object_get_type(node);

	switch (type) {
	case json_type_boolean:
		value->type = VKC_BOOL;
		value->value.b = json_object_get_boolean(node);
		break;
	case json_type_double:
		value->type = VKC_DOUBLE;
		value->value.d = json_object_get_double(node);
		break;
	case json_type_int:
		value->type = VKC_INT;
		value->value.i = json_object_get_int(node);
		break;
	case json_type_string:
		value->type = VKC_STRING;
		str = json_object_get_string(node);
		if (!str)
			goto err;

		value->value.s = strdup(str);
		break;
	case json_type_array:
	case json_type_object:
		log_error("Field %s must have simple type", key);
		return -EINVAL;
		break;
	case json_type_null:
		return -EINVAL;
		break;
	}

	return 0;

err:
	return -1;
}

static int parse_vconf_data(json_object *root, struct rule *r)
{
	struct vconf_key_changed_event_data *evd;
	int ret;

	evd = calloc(1, sizeof(*evd));
	if (!evd)
		return -ENOMEM;

	(void) get_config_field(root, "key_name", &evd->key_name, json_type_string);
	ret = parse_vconf_field(root, "oldval", &evd->oldval);
	r->match_oldval = ret >= 0;
	ret = parse_vconf_field(root, "newval", &evd->newval);
	r->match_newval = ret >= 0;

	r->ev_data = evd;
	r->type = RULE_TYPE_VCONF;
	return 0;
}

static int parse_dbus_data(json_object *root, struct rule *r)
{
	struct dbus_signal_event_data *evd;

	evd = calloc(1, sizeof(*evd));
	if (!evd)
		return -ENOMEM;

	r->match = NULL;

	(void) get_config_field(root, "id", &evd->event_id, json_type_string);
	(void) get_config_field(root, "match", &r->match, json_type_object);

	r->ev_data = evd;
	r->type = RULE_TYPE_DBUS;
	return 0;
}

static void rule_cleanup(struct rule *r)
{
	free(r->ev_data);
}

static void clean_rules(struct activation_decision_maker *dm)
{
	struct rule *r, *next;

	list_for_each_entry_safe(r, next, &dm->rules, node) {
		list_del_init(&r->node);
		rule_cleanup(r);
		free(r);
	}
}

static int add_rule(struct activation_decision_maker *dm, json_object *root)
{
	int ret;
	struct rule *r;

	r = calloc(1, sizeof(*r));
	if (!r) {
		log_error("Could not allocate rule object");
		return -ENOMEM;
	}

	ret = get_config_field(root, "event", &r->event, json_type_string);
	if (ret < 0)
		goto out;

	log_debug("Adding rule for event %s", r->event);

	if (strcmp(r->event, VCONF_KEY_CHANGED_EVENT_ID) == 0) {
		parse_vconf_data(root, r);
	} else if (strcmp(r->event, DBUS_SIGNAL_EVENT_ID) == 0) {
		parse_dbus_data(root, r);
	} else {
		log_warning("Event type %s is not supported by activationd module\n", r->event);
		goto out;
	}

	ret = get_config_field(root, "action", &r->action, json_type_string);
	if (ret < 0)
		goto out;

	ret = get_config_field(root, "target", &r->target, json_type_string);
	if (ret < 0)
		goto out;

	list_add_tail(&r->node, &dm->rules);

	return 0;

out:
	rule_cleanup(r);
	free(r);
	return ret;
}

static int activation_init(struct epc_event_handler *handler, struct epc_config *config)
{
	struct activation_decision_maker *dm = container_of(handler,
			struct activation_decision_maker, eh);
	int len;
	int i;
	int ret;
	json_object *array, *obj;

	INIT_LIST_HEAD(&dm->rules);

	if (config == NULL)
		return 0;

	if (!json_object_object_get_ex(config->root, "rules", &array)) {
		log_error("There are no 'rules' array");
		return -EINVAL;
	}

	if (!json_object_is_type(array, json_type_array)) {
		log_error("Config value is not an array");
		return -EINVAL;
	}

	len = json_object_array_length(array);
	for (i = 0; i < len; ++i) {
		obj = json_object_array_get_idx(array, i);
		ret = add_rule(dm, obj);
		if (ret < 0) {
			log_error("Could not add rule");
			goto cleanup;
		}
	}

	return 0;

cleanup:
	clean_rules(dm);
	return ret;
};

static void activation_cleanup(struct epc_event_handler *handler)
{
	struct activation_decision_maker *dm = container_of(handler,
			struct activation_decision_maker, eh);

	clean_rules(dm);
}

static struct activation_decision_maker activation_dm = {
	.eh = {
		.name = MODULE_NAME,
		.init = activation_init,
		.cleanup = activation_cleanup,
		.event_match = activation_event_match,
		.handle_event = activation_make_decision,

		.node = LIST_HEAD_INIT(activation_dm.eh.node),
	},
};

EPC_EVENT_HANDLER_REGISTER(activation_dm.eh, activation_eh,
		EPC_MODULE_TYPE_DECISION_MAKER)
