/*
 * This file is part of unit_controld.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <malloc.h>
#include <errno.h>
#include <string.h>
#include <libsyscommon/dbus-system.h>

#include "decision_made_event.h"
#include "action.h"
#include "event_processor.h"
#include "log.h"
#include "service.h"
#include "common.h"
#include "database.h"
#include "json-config.h"
#include "systemd_dbus.h"

#include "unit_control_event.h"

#define MODULE_NAME "unit_control_decision_maker"

struct unit_control_decision_maker {
	struct epc_event_handler eh;
	struct list_head whitelist;
};

struct whitelist_entry {
	char *unit;
	struct list_head node;
};

struct unit_info {
	char *unit;
	struct list_head node;
};

static int unit_control_event_match(struct epc_event_handler *handler,
		struct epc_event *ev)
{
	if (strcmp(ev->type->name, UC_EVENT_ID) == 0)
			return 1;
	return 0;
}

static int list_unit_by_pattern(const char *pattern, GVariant **reply)
{
	assert(pattern);

	GVariant *result;
	char *patterns[2] = {};
	char *unit_states[] = {
		"loaded",
		NULL
	};

	patterns[0] = (char *)pattern;
	result = dbus_handle_method_sync_with_reply_var(
			SYSTEMD_SERVICE,
			SYSTEMD_OBJ,
			SYSTEMD_MANAGER_INTERFACE,
			"ListUnitsByPatterns",
			g_variant_new("(asas)", (char **)unit_states, (char **)patterns));

	if (!result) {
		log_error("Failed to issue ListUnitsByPatterns() call.");
		return -1;
	}

	*reply = result;

	return 0;
}

static void parse_unit_info(GVariant *m, char **name)
{
	assert(m);

	char *id;

	g_variant_get(
		m,
		"(ssssssouso)",
		&id,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL
		);

	*name = id;
}

static int parse_unit_list(GVariant *reply, struct list_head *units)
{
	assert(reply);
	assert(units);

	char *name;
	struct unit_info *u;
	GVariantIter iter;
	GVariant *child;

	g_variant_get(reply, "a(ssssssouso)", &iter);

	while ((child = g_variant_iter_next_value(&iter))) {
		parse_unit_info(child, &name);
		u = calloc(1, sizeof(*u));

		if (!u) {
			log_error("Could not allocate unit_info object");
			return -ENOMEM;
		}

		u->unit = strdup(name);
		if (!u->unit) {
			free(u);
			return -ENOMEM;
		}

		list_add_tail(&u->node, units);
	}

	return 0;
}

static void unit_control_executed(struct decision_made_event *dm_ev, int result)
{
	struct unit_control_event *event = to_unit_control_event(dm_ev->reason);

	if (result < 0)
		event->err_code = result;

	event->jobs_cnt--;
	if (event->jobs_cnt == 0) {
		if (event->err_code < 0) {
			g_dbus_method_invocation_return_value(event->m, g_variant_new("(s)", "error"));
		} else {
			g_dbus_method_invocation_return_value(event->m, g_variant_new("(s)", "ok"));
		}
	}

	epc_event_unref(&event->event);
}

static int unit_control_execute(const char *unit, const char *command, struct epc_event *ev)
{
	assert(unit);
	assert(command);
	assert(ev);

	struct dm_event_data ev_data = {
		.reason = ev,
		.who_made = MODULE_NAME,
		.on_finished = unit_control_executed,
	};
	struct epc_event *new_ev;
	int ret;
	struct unit_control_event *event = to_unit_control_event(ev);

	ev_data.action = EPC_ACTION_UNIT_START_ID;
	ret = epc_object_new(&ev_data.action_data);
	if (ret < 0) {
		log_error_errno(ret, "could not create data object for action - %m");
		goto out;
	}

	ret = epc_fill_for_unit_start(ev_data.action_data, unit);
	if (ret) {
		log_error_errno(ret, "unable to create event data - %m");
		goto out;
	}

	if (!strcmp(command, "start"))
		epc_fill_for_unit_action(ev_data.action_data, "StartUnit");
	else if (!strcmp(command, "stop"))
		epc_fill_for_unit_action(ev_data.action_data, "StopUnit");
	else if (!strcmp(command, "restart"))
		epc_fill_for_unit_action(ev_data.action_data, "RestartUnit");
	else {
		ret = -EINVAL;
		log_error_errno(ret, "could not dispatch action: %s - %m", command);
		goto out;
	}

	ret = epc_event_create(DECISION_MADE_EVENT_ID, &ev_data, &new_ev);
	if (ret) {
		log_error("Unable to create event");
		goto out;
	}

	epc_event_ref(&event->event);
	event->jobs_cnt++;

	ret = event_processor_report_event(new_ev);
	epc_event_unref(new_ev);

	if (ret) {
		event->jobs_cnt--;
		epc_event_unref(&event->event);
		log_error_errno(ret, "unable to report event - %m");
		goto out;
	}

out:
	epc_object_unref(ev_data.action_data);
	return ret;
}

static int unit_control_query_and_execute(const char *pattern, const char *command, struct epc_event *ev)
{
	assert(pattern);
	assert(command);
	assert(ev);

	int ret, error_code = 0;
	struct list_head unit_list;
	struct unit_info *u, *next;
	struct unit_control_event *event = to_unit_control_event(ev);
	GVariant *reply;

	INIT_LIST_HEAD(&unit_list);

	ret = list_unit_by_pattern(pattern, &reply);
	if (ret < 0)
		return ret;
	ret = parse_unit_list(reply, &unit_list);
	if (ret < 0)
		goto cleanup;

	list_for_each_entry_safe(u, next, &unit_list, node) {
		log_debug("execute action %s on unit %s", command, u->unit);
		ret = unit_control_execute(u->unit, command, ev);
		if (ret < 0)
			error_code = ret;
		free(u->unit);
		list_del(&u->node);
	}

	if (error_code != 0)
		return error_code;

	/* nothing more to do */
	if (event->jobs_cnt == 0) {
		g_dbus_method_invocation_return_value(event->m, g_variant_new("(s)", "error"));
	}

	return 0;
cleanup:
	list_for_each_entry_safe(u, next, &unit_list, node) {
		free(u->unit);
		list_del(&u->node);
	}
	return ret;
}

static int match_unit(const char *unit_name, const char *unit_whitelist_rule)
{
	assert(unit_name);
	assert(unit_whitelist_rule);

	char *rule_suf, *unit_suf;

	rule_suf = strrchr(unit_whitelist_rule, '.');
	if (rule_suf && rule_suf != unit_whitelist_rule && *(rule_suf - 1) == '@') {
		unit_suf = strrchr(unit_name, '.');
		if (unit_suf && !strncmp(unit_name, unit_whitelist_rule, rule_suf - unit_whitelist_rule)
				&& !strcmp(unit_suf, rule_suf))
			return 1;

		return 0;
	}

	if (!strcmp(unit_name, unit_whitelist_rule))
		return 1;

	return 0;
}

/*
 * Assuming method is not 'start', in both cases:
 *	Request:				Whitelist entry:
 *	---------------------------------------
 *	name@instance.service	name@.service
 *	name@.service			name@instance.service
 *
 * the right action to execute is equivalent to:
 * systemctl stop|restart name@instance.service
 *
 * So in any case we need to pick the most precise name.
 * Comparing string lengths seems quick and easy way to do this.
 *
 * Next thing is, if the proper name is a wildcard, e.g. name@.service,
 * the way of sending a method call is to add a '*' to the service name:
 * name@*.service
 *
 * User needs to free returned pointer.
 * */
static char *construct_proper_service_name(char *requested_name, char *whitelist_name)
{
	assert(requested_name);
	assert(whitelist_name);

	char *name, *ret, *at;
	int buf_size;

	if (strlen(requested_name) > strlen(whitelist_name))
		name = requested_name;
	else
		name = whitelist_name;

	at = strstr(name, "@.");

	/* add '*' if wildcard */
	if (at) {
		at++; /* move the pointer after '@' */
		buf_size = strlen(name) + 2;
		ret = calloc(1, buf_size); /* one for \0 and one for additional asterisk*/
		if (!ret)
			return NULL;
		strncpy(ret, name, at - name);
		snprintf(ret + (at - name), buf_size-(at - name), "*%s", at);
	} else {
		ret = strdup(name);
	}

	return ret;
}

static bool is_wildcard(char *service_name)
{
	char *c;

	c = strchr(service_name, '*');
	return c && *c != '\0';
}

static int unit_control_make_decision(struct epc_event_handler *handler)
{
	assert(handler);

	struct epc_event *event = pop_epc_event(&handler->event_queue);
	struct unit_control_decision_maker *dm = container_of(handler,
			struct unit_control_decision_maker, eh);
	struct unit_control_event *ev = to_unit_control_event(event);
	bool matched = false;
	int error_code = 0;
	struct whitelist_entry *w;
	_cleanup_free_ char *service_name = NULL;

	/* cannot use start with wildcards */
	if (!strcmp(ev->method, "start") && strstr(ev->unit, "@.")) {
		error_code = -ENOTSUP;
		goto cleanup;
	}

	list_for_each_entry(w, &dm->whitelist, node) {
		if (!match_unit(ev->unit, w->unit))
			continue;

		matched = true;

		service_name = construct_proper_service_name(ev->unit, w->unit);

		if (!service_name) {
			error_code = -ENOMEM;
			break;
		}

		log_debug("execute action %s on pattern %s", ev->method, service_name);

		if (!strcmp(ev->method, "start") || !is_wildcard(service_name)) /* no need to query when starting or for single unit*/
			error_code = unit_control_execute(service_name, ev->method, event);
		else
			error_code = unit_control_query_and_execute(service_name, ev->method, event);

		if (error_code < 0) {
			log_error_errno(error_code, "action %s on unit %s failed (%m)", ev->method, service_name);
			goto cleanup;
		}
	}

	if (matched != true)
		error_code = -EPERM;

cleanup:
	if (error_code != 0) {
		g_dbus_method_invocation_return_value(ev->m, g_variant_new("(s)", "error"));
	}
	epc_event_unref(event);
	return 0;
}

static void cleanup_whitelist_entry(struct whitelist_entry *w)
{
	assert(w);
	free(w->unit);
	free(w);
}

static void cleanup_whitelist(struct unit_control_decision_maker *dm)
{
	assert(dm);

	struct whitelist_entry *w, *next;

	list_for_each_entry_safe(w, next, &dm->whitelist, node) {
		list_del_init(&w->node);
		cleanup_whitelist_entry(w);
	}
}

static int add_whitelist_entry(struct unit_control_decision_maker *dm, const char *unit)
{
	assert(dm);
	assert(unit);

	struct whitelist_entry *w, *next;

	/* check for duplicates and generalizations */
	list_for_each_entry_safe(w, next, &dm->whitelist, node) {
		if (match_unit(unit, w->unit)) {
			log_debug("not adding %s to the whitelist - duplicate/more general rule %s exists", unit, w->unit);
			return 0;
		}
		if (match_unit(w->unit, unit)) {
			log_debug("add %s to the whitelist - more general rule than %s - removing %s", unit, w->unit, w->unit);
			/* remove w */
			list_del_init(&w->node);
			cleanup_whitelist_entry(w);
		}
	}

	w = calloc(1, sizeof(*w));
	if (!w) {
		log_error("could not allocate whitelist entry object");
		return -ENOMEM;
	}

	log_debug("adding whitelist entry for unit %s", unit);

	w->unit = strdup(unit);
	list_add_tail(&w->node, &dm->whitelist);
	return 0;
}

static int unit_control_init(struct epc_event_handler *handler, struct epc_config *config)
{
	assert(handler);

	struct unit_control_decision_maker *dm = container_of(handler,
			struct unit_control_decision_maker, eh);
	int len;
	int i;
	int ret;
	json_object *array, *obj;

	INIT_LIST_HEAD(&dm->whitelist);

	if (config == NULL)
		return 0;

	if (!json_object_object_get_ex(config->root, "whitelist", &array)) {
		log_error("there is no 'whitelist' array");
		return -EINVAL;
	}

	if (!json_object_is_type(array, json_type_array)) {
		log_error("config value is not an array");
		return -EINVAL;
	}

	len = json_object_array_length(array);
	for (i = 0; i < len; ++i) {
		obj = json_object_array_get_idx(array, i);
		ret = add_whitelist_entry(dm, json_object_get_string(obj));
		if (ret < 0) {
			log_error_errno(ret, "could not add whitelist entry - %m");
			goto cleanup;
		}
	}
	return 0;

cleanup:
	cleanup_whitelist(dm);
	return ret;
};

static void unit_control_cleanup(struct epc_event_handler *handler)
{
	assert(handler);

	struct unit_control_decision_maker *dm = container_of(handler,
			struct unit_control_decision_maker, eh);

	cleanup_whitelist(dm);
}

static struct unit_control_decision_maker unit_control_dm = {
	.eh = {
		.name = MODULE_NAME,
		.init = unit_control_init,
		.cleanup = unit_control_cleanup,
		.event_match = unit_control_event_match,
		.handle_event = unit_control_make_decision,
		.node = LIST_HEAD_INIT(unit_control_dm.eh.node),
	},
};

EPC_EVENT_HANDLER_REGISTER(unit_control_dm.eh, unit_control_eh,
		EPC_MODULE_TYPE_DECISION_MAKER)
