/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <getopt.h>
#include <json-c/json.h>
#include <stdio.h>
#include <poll.h>
#include <systemd/sd-bus.h>
#include <systemd/sd-event.h>
#include <signal.h>
#include <syslog.h>
#include <unistd.h>

#include "log.h"
#include "module.h"
#include "systemd_dbus.h"
#include "epc-config.h"
#include "common.h"
#include "epc-glib.h"
#include "json-config.h"
#include "database.h"

#define EPC_GLIB_MAINLOOP

static int terminate = 0;

enum epc_running_modes {
	EPC_MODE_NORMAL,
	EPC_MODE_CTL,
};

static int running_mode = EPC_MODE_NORMAL;

static int signal_handler(sd_event_source *s,
				   const struct signalfd_siginfo *si,
				   void *userdata)
{
	sd_event *loop = userdata;

	sd_event_exit(loop, 0);
	terminate = 1;
	return 0;
}

static void print_help(const char *name)
{
	printf("Usage: %s [OPTIONS]\n"
		   "Options: \n"
		   "  --log-level <level>            Set log level to <level>\n"
		   "  -D                             Set log level to debug\n"
		   "  -c, --config-file <file>       Read epc configuration from <file>\n",
		   name);

	switch (running_mode) {
	case EPC_MODE_NORMAL:
		printf("  -d, --disable-module <module>  Disable <module>\n"
			   "  --systemd-private              Use private connection to systemd\n"
			   "  --no-action                    Do not execute any action\n");
		break;

	case EPC_MODE_CTL:
		/* No command line arguments for now */;
		break;
	}

	printf("  -h, --help                     Display this help and exit\n");
}

static int parse_argv(int ac, char *av[])
{
	int c, r;

	#define CHECK_MODE(m) \
		if (running_mode != m) { \
			print_help(av[0]); \
			return -EINVAL; \
		}

	enum {
		ARG_LOG_LEVEL = 0x100,
		ARG_NO_ACTION,
		ARG_PRIVATE,
	};

	static const struct option options[] = {
		{"log-level", required_argument, NULL, ARG_LOG_LEVEL},
		{"disable-module", required_argument, NULL, 'd'},
		{"systemd-private", no_argument, NULL, ARG_PRIVATE},
		{"no-action", no_argument, NULL, ARG_NO_ACTION},
		{"config-file", required_argument, NULL, 'c'},
		{"help", no_argument, NULL, 'h'},
		{}
	};

	struct epc_module *module;
	const char *custom_config_file = EPC_CUSTOM_CONFIG_FILE;

	if (strcmp(basename(av[0]), "actdctl") == 0)
		running_mode = EPC_MODE_CTL;

	while ((c = getopt_long(ac, av, "Dc:d:h", options, NULL)) >= 0) {
		switch (c) {
		case ARG_LOG_LEVEL:
			r = log_parse_level_name(optarg);
			if (r < 0) {
				fprintf(stderr, "Unknown log level: \"%s\"\n", optarg);
				return r;
			}
			log_set_max_level(r);
			break;

		case 'D':
			log_set_max_level(LOG_DEBUG);
			break;

		case 'c':
			custom_config_file = optarg;
			break;

		case 'h':
			print_help(av[0]);
			exit(0);

		case 'd':
			CHECK_MODE(EPC_MODE_NORMAL);

			module = epc_get_module_by_name(optarg);
			if (module) {
				module->disabled = 1;
				log_debug("Disabled %s module", optarg);
			} else {
				log_warning("Could not find %s module", optarg);
			}

			break;

		case ARG_NO_ACTION:
			CHECK_MODE(EPC_MODE_NORMAL);

			epc_disable_module_category(EPC_MODULE_TYPE_ACTION);
			break;

		case ARG_PRIVATE:
			epc_set_systemd_private(true);
			break;

		default:
			print_help(av[0]);
			return -EINVAL;
		}
	}

	return epc_load_config(custom_config_file);
	#undef CHECK_MODE
}

int main(int ac, char *av[])
{
	int rc;
	sd_bus *bus = NULL;
	sd_event *loop;
	sigset_t ss;
	struct epc_config *modules_config = NULL;
	char err[512];

	rc = epc_modules_load();
	if (rc < 0) {
		log_error("Failed to load modules");
		return -1;
	}

	rc = parse_argv(ac, av);
	if (rc < 0) {
		if (strerror_r(-rc, err, sizeof err))
			fprintf(stderr, "Failed to parse command line options: %s\n", err);
		return -1;
	}

	rc = epc_acquire_systemd_bus(&bus);
	if (rc < 0) {
		log_error_errno(rc, "Failed to acquire the default system bus connection: %m");
		return -1;
	}

	rc = sd_event_new(&loop);
	if (rc < 0) {
		log_error_errno(rc, "Failed to allocate the event loop.");
		return -1;
	}

	/* If there is way to register a catch-all handler */
	/* sd_bus_attach_event(bus, loop, SD_EVENT_PRIORITY_NORMAL); */
	sigemptyset(&ss);
	sigaddset(&ss, SIGINT);
	sigaddset(&ss, SIGTERM);
	rc = sigprocmask(SIG_BLOCK, &ss, NULL);
	if (rc < 0) {
		log_error("Failed to change signal mask %d.", rc);
		return -1;
	}

	rc = sd_event_add_signal(loop, NULL, SIGINT, signal_handler, loop);
	if (rc < 0) {
		log_error("Failed to register SIGINT handler %d.", rc);
		return -1;
	}

	rc = sd_event_add_signal(loop, NULL, SIGTERM, signal_handler, loop);
	if (rc < 0) {
		log_error("Failed to register SIGTERM handler %d.", rc);
		return -1;
	}

	epc_get_modules_config(&modules_config);

	switch (running_mode) {
	case EPC_MODE_NORMAL:
		rc = epc_modules_init(loop, modules_config);
		if (rc < 0) {
			log_error("Failed to initialize modules %d.", rc);
			return -1;
		}

		printf("Hello world!\n");

#ifdef EPC_GLIB_MAINLOOP
		rc = epc_g_main_loop_wrapper(loop);
#else
		rc = sd_event_loop(loop);
#endif /* EPC_GLIB_MAINLOOP */

		if (rc < 0)
			log_error_errno(rc, "Failed to run main loop: %m");
		break;

	case EPC_MODE_CTL:
	{
		struct epc_object *result = NULL;
		char *buf = NULL;

		epc_disable_module_category(EPC_MODULE_TYPE_UNCATEGORIZED);
		epc_disable_module_category(EPC_MODULE_TYPE_EVENT);
		epc_disable_module_category(EPC_MODULE_TYPE_ACTION_EXECUTOR);
		epc_disable_module_category(EPC_MODULE_TYPE_ACTION);
		epc_disable_module_category(EPC_MODULE_TYPE_DECISION_MAKER);
		epc_disable_module_category(EPC_MODULE_TYPE_LISTENER);

		rc = epc_modules_init(loop, modules_config);
		if (rc < 0) {
			log_error("Failed to initialize modules %d.", rc);
			return -1;
		}

		rc = database_dump(&result);
		if (rc < 0) {
			log_error("Failed to dump database");
			return -1;
		}

		rc = epc_object_to_string(&buf, result, 0);
		epc_object_unref(result);
		if (rc < 0) {
			log_error("Unable to convert epc object to string");
			return -1;
		}

		printf("%s\n", buf);

		free(buf);
		break;
	}
	}

	epc_modules_cleanup();
	epc_config_cleanup_global();

	sd_bus_close(bus);

	rc = epc_call_at_exit();

	epc_modules_unload();

	return rc;
}
