/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "common.h"
#include "database.h"
#include "action_executed_event.h"

static int allocate_ae_event(struct epc_event_type *type,
							 void *data, struct epc_event **ev)
{
	struct action_executed_event *ae_ev;
	struct ae_event_data *ae_ev_data = data;
	int ret;

	ae_ev = calloc(1, sizeof(*ae_ev));
	if (!ae_ev)
		return -ENOMEM;

	ret = epc_event_init_internal(type, &ae_ev->event);
	if (ret)
		goto free_ae_ev;

	ae_ev->action = strdup(ae_ev_data->action);
	ae_ev->action_impl = strdup(ae_ev_data->action_impl);

	if (!ae_ev->action || !ae_ev->action_impl) {
		ret = -ENOMEM;
		goto cleanup_strs;
	}

	ae_ev->action_log = epc_object_ref(ae_ev_data->action_log);
	ae_ev->result  = ae_ev_data->result;

	ae_ev->reason = ae_ev_data->reason;
	if (ae_ev->reason)
		epc_event_ref(ae_ev->reason);

	*ev = &ae_ev->event;
	return 0;
cleanup_strs:
	free(ae_ev->action);
	free(ae_ev->action_impl);

	epc_event_cleanup_internal(&ae_ev->event);
free_ae_ev:
	free(ae_ev);

	return ret;
}

static int deserialize_ae_event(struct epc_event_type *type,
								struct epc_object *data, struct epc_event **ev)
{
	int ret = -EINVAL;
	struct ae_event_data ae_ev_data;
	struct epc_object *obj;

	memset(&ae_ev_data, 0, sizeof(ae_ev_data));

	list_for_each_entry(obj, &data->val.children, node) {
		if ((obj->type == TYPE_OID) &&
			(strcmp(AE_EV_REASON, obj->key) == 0)) {

			ret = database_get_event(&obj->val.oid, &ae_ev_data.reason);
			if (ret)
				goto finish;

		} else if ((obj->type == TYPE_STRING) &&
				   (strcmp(AE_EV_ACTION, obj->key) == 0)) {

			ae_ev_data.action = obj->val.s;

		} else if ((obj->type == TYPE_STRING) &&
				   (strcmp(AE_EV_ACTION_IMPL, obj->key) == 0)) {

			ae_ev_data.action_impl = obj->val.s;

		} else if ((obj->type == TYPE_OBJECT) &&
			(strcmp(AE_EV_ACTION_LOG, obj->key) == 0)) {

			ae_ev_data.action_log = obj;
		} else if ((obj->type == TYPE_INT) &&
			(strcmp(AE_EV_RESULT, obj->key) == 0)) {

			ae_ev_data.result = obj->val.i;
		}
	}

	if (!ae_ev_data.reason ||
		!ae_ev_data.action ||
		!ae_ev_data.action_impl ||
		!ae_ev_data.action_log) {
			ret = -EINVAL;
			goto finish;
		}

	ret = allocate_ae_event(type, &ae_ev_data, ev);
	if (ret < 0)
		goto finish;

	ret = epc_event_deserialize_internal(data, type, *ev);
	if (ret < 0) {
		struct action_executed_event *ae_ev =
			to_action_executed_event(*ev);
		epc_event_unref(ae_ev->reason);
		epc_event_unref(&ae_ev->event);
		goto finish;
	}

	ret = 0;
finish:
	return ret;
}
TEST_ENTRY_POINT(deserialize_ae_event);

static void ae_event_release(struct epc_event *ev)
{
	struct action_executed_event *ae_ev =
		to_action_executed_event(ev);

	if (ae_ev->reason)
		epc_event_unref(ae_ev->reason);
	free(ae_ev->action);
	free(ae_ev->action_impl);
	epc_object_unref(ae_ev->action_log);
	epc_event_cleanup_internal(&ae_ev->event);
	free(ae_ev);
}

static void ae_event_serialize(struct epc_event *ev, struct epc_object *out)
{
	struct action_executed_event *ae_ev =
		to_action_executed_event(ev);

	assert(database_is_oid_valid(&ae_ev->reason->oid));

	epc_event_serialize_internal(ev, out);
	epc_object_append_oid(out, AE_EV_REASON, &ae_ev->reason->oid);
	epc_object_append_string(out, AE_EV_ACTION, ae_ev->action);
	epc_object_append_string(out, AE_EV_ACTION_IMPL, ae_ev->action_impl);
	epc_object_append_object(out, AE_EV_ACTION_LOG, ae_ev->action_log);
	epc_object_append_int(out, AE_EV_RESULT, ae_ev->result);
}
TEST_ENTRY_POINT(ae_event_serialize);

static struct epc_event_type action_executed_event_type = {
	.name = ACTION_EXECUTED_EVENT_ID,
	.default_ops = {
		.release = ae_event_release,
		.serialize = ae_event_serialize,
	},
	.allocate_event = allocate_ae_event,
	.deserialize_event = deserialize_ae_event,
	.node = LIST_HEAD_INIT(action_executed_event_type.node),
};

EPC_EVENT_TYPE_REGISTER(action_executed_event_type, action_executed_et)
