/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "dbus_signal_event.h"

static int allocate_dbus_signal_event(struct epc_event_type *type,
							 void *data, struct epc_event **ev)
{
	struct dbus_signal_event *ds_ev;
	struct dbus_signal_event_data *ds_ev_data = data;
	int ret;

	ds_ev = calloc(1, sizeof(*ds_ev));
	if (!ds_ev)
		return -ENOMEM;

	ret = epc_event_init_internal(type, &ds_ev->event);
	if (ret)
		goto cleanup;

	ds_ev->event_time = ds_ev_data->event_time;
	ds_ev->event_id = ds_ev_data->event_id;
	ds_ev->sender = ds_ev_data->sender;
	ds_ev->interface = ds_ev_data->interface;
	ds_ev->member = ds_ev_data->member;
	ds_ev->path = ds_ev_data->path;
	ds_ev->path_namespace = ds_ev_data->path_namespace;
	ds_ev->destination = ds_ev_data->destination;
	ds_ev->match_params = ds_ev_data->match_params;
	ds_ev->signal_params = ds_ev_data->signal_params;

	*ev = &ds_ev->event;
	return 0;
cleanup:
	free(ds_ev);

	return ret;
}

static int deserialize_dbus_signal_event(struct epc_event_type *type,
								struct epc_object *data, struct epc_event **ev)
{
	int ret = -EINVAL;
	struct dbus_signal_event_data ds_ev_data;
	struct epc_object *obj;
	memset(&ds_ev_data, 0, sizeof(ds_ev_data));

	list_for_each_entry(obj, &data->val.children, node) {
		switch (obj->type) {
		case TYPE_TIME_T:
			if (!strcmp(DS_EV_TIME, obj->key))
				ds_ev_data.event_time = obj->val.time;
			break;
		case TYPE_STRING:
			if (!strcmp(DS_EV_ID, obj->key))
				ds_ev_data.event_id = strdup(obj->val.s);
			else if (!strcmp(DS_EV_SENDER, obj->key))
				ds_ev_data.sender = strdup(obj->val.s);
			else if (!strcmp(DS_EV_INTERFACE, obj->key))
				ds_ev_data.interface = strdup(obj->val.s);
			else if (!strcmp(DS_EV_MEMBER, obj->key))
				ds_ev_data.member = strdup(obj->val.s);
			else if (!strcmp(DS_EV_PATH, obj->key))
				ds_ev_data.path = strdup(obj->val.s);
			else if (!strcmp(DS_EV_PATH_NAMESPACE, obj->key))
				ds_ev_data.path_namespace = strdup(obj->val.s);
			else if (!strcmp(DS_EV_DESTINATION, obj->key))
				ds_ev_data.destination = strdup(obj->val.s);
			break;
		case TYPE_OBJECT:
			if (!strcmp(DS_EV_PARAMS, obj->key)) {
				epc_object_ref(obj);
				ds_ev_data.match_params = obj;
			}
			break;
		}
	}

	ret = allocate_dbus_signal_event(type, &ds_ev_data, ev);
	if (ret < 0)
		goto finish;

	ret = epc_event_deserialize_internal(data, type, *ev);
	if (ret < 0) {
		struct dbus_signal_event *ds_ev =
			to_dbus_signal_event(*ev);
		free(ds_ev);
		goto finish;
	}
	ret = 0;
finish:
	return ret;
}

static void dbus_signal_event_release(struct epc_event *ev)
{
	struct dbus_signal_event *ds_ev =
		to_dbus_signal_event(ev);

	free(ds_ev->event_id);
	free(ds_ev->sender);
	free(ds_ev->interface);
	free(ds_ev->member);
	free(ds_ev->path);
	free(ds_ev->path_namespace);
	free(ds_ev->destination);
	if (ds_ev->match_params)
		epc_object_destroy(ds_ev->match_params);

	if (ds_ev->signal_params)
		g_variant_unref(ds_ev->signal_params);

	epc_event_cleanup_internal(&ds_ev->event);
	free(ds_ev);
}

static void dbus_signal_event_serialize(struct epc_event *ev, struct epc_object *out)
{
	struct dbus_signal_event *ds_ev =
		to_dbus_signal_event(ev);
	epc_event_serialize_internal(ev, out);
	epc_object_append_time_t(out, DS_EV_TIME, ds_ev->event_time);
	epc_object_append_string(out, DS_EV_ID, ds_ev->event_id);
	if (ds_ev->sender)
		epc_object_append_string(out, DS_EV_SENDER, ds_ev->sender);
	if (ds_ev->interface)
		epc_object_append_string(out, DS_EV_INTERFACE, ds_ev->interface);
	if (ds_ev->member)
		epc_object_append_string(out, DS_EV_MEMBER, ds_ev->member);
	if (ds_ev->path)
		epc_object_append_string(out, DS_EV_PATH, ds_ev->path);
	if (ds_ev->path_namespace)
		epc_object_append_string(out, DS_EV_PATH_NAMESPACE, ds_ev->path_namespace);
	if (ds_ev->destination)
		epc_object_append_string(out, DS_EV_DESTINATION, ds_ev->destination);
	if (ds_ev->match_params)
		epc_object_append_object(out, DS_EV_PARAMS, ds_ev->match_params);
}

static struct epc_event_type dbus_signal_event_type = {
	.name = DBUS_SIGNAL_EVENT_ID,
	.default_ops = {
		.release = dbus_signal_event_release,
		.serialize = dbus_signal_event_serialize,
	},
	.allocate_event = allocate_dbus_signal_event,
	.deserialize_event = deserialize_dbus_signal_event,
	.node = LIST_HEAD_INIT(dbus_signal_event_type.node),
};

EPC_EVENT_TYPE_REGISTER(dbus_signal_event_type, dbus_signal_et)
