/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "common.h"
#include "database.h"
#include "decision_made_event.h"

static int allocate_dm_event(struct epc_event_type *type,
							 void *data, struct epc_event **ev)
{
	struct decision_made_event *dm_ev;
	struct dm_event_data *dm_ev_data = data;
	int ret;

	dm_ev = calloc(1, sizeof(*dm_ev));
	if (!dm_ev)
		return -ENOMEM;

	ret = epc_event_init_internal(type, &dm_ev->event);
	if (ret)
		goto free_dm_ev;

	dm_ev->who_made = strdup(dm_ev_data->who_made);
	dm_ev->action = strdup(dm_ev_data->action);
	dm_ev->on_finished = dm_ev_data->on_finished;

	if (!dm_ev->who_made || !dm_ev->action) {
		ret = -ENOMEM;
		goto cleanup_strs;
	}

	dm_ev->action_data = epc_object_ref(dm_ev_data->action_data);

	dm_ev->reason = dm_ev_data->reason;
	epc_event_ref(dm_ev->reason);

	*ev = &dm_ev->event;
	return 0;
cleanup_strs:
	free(dm_ev->who_made);
	free(dm_ev->action);

	epc_event_cleanup_internal(&dm_ev->event);
free_dm_ev:
	free(dm_ev);

	return ret;
}

static int deserialize_dm_event(struct epc_event_type *type,
								struct epc_object *data, struct epc_event **ev)
{
	int ret = -EINVAL;
	struct dm_event_data dm_ev_data;
	struct epc_object *obj;

	memset(&dm_ev_data, 0, sizeof(dm_ev_data));

	list_for_each_entry(obj, &data->val.children, node) {
		if ((obj->type == TYPE_OID) &&
			(strcmp(DM_EV_REASON, obj->key) == 0)) {

			ret = database_get_event(&obj->val.oid, &dm_ev_data.reason);
			if (ret)
				goto finish;

		} else if ((obj->type == TYPE_STRING) &&
				   (strcmp(DM_EV_WHO, obj->key) == 0)) {

			dm_ev_data.who_made = obj->val.s;

		} else if ((obj->type == TYPE_STRING) &&
				   (strcmp(DM_EV_ACTION, obj->key) == 0)) {

			dm_ev_data.action = obj->val.s;

		} else if ((obj->type == TYPE_OBJECT) &&
			(strcmp(DM_EV_ACTION_DATA, obj->key) == 0)) {

			dm_ev_data.action_data = obj;
		}
	}

	if (!dm_ev_data.reason ||
		!dm_ev_data.who_made ||
		!dm_ev_data.action ||
		!dm_ev_data.action_data) {
			ret = -EINVAL;
			goto finish;
		}

	ret = allocate_dm_event(type, &dm_ev_data, ev);
	if (ret < 0)
		goto finish;

	ret = epc_event_deserialize_internal(data, type, *ev);
	if (ret < 0) {
		struct decision_made_event *dm_ev =
			to_decision_made_event(*ev);
		epc_event_unref(dm_ev->reason);
		epc_event_unref(&dm_ev->event);
		goto finish;
	}

	ret = 0;
finish:
	return ret;
}
TEST_ENTRY_POINT(deserialize_dm_event);

static void dm_event_release(struct epc_event *ev)
{
	struct decision_made_event *dm_ev =
		to_decision_made_event(ev);

	free(dm_ev->who_made);
	free(dm_ev->action);
	epc_object_unref(dm_ev->action_data);
	epc_event_unref(dm_ev->reason);
	epc_event_cleanup_internal(&dm_ev->event);
	free(dm_ev);
}

static void dm_event_serialize(struct epc_event *ev, struct epc_object *out)
{
	struct decision_made_event *dm_ev =
		to_decision_made_event(ev);

	assert(database_is_oid_valid(&dm_ev->reason->oid));

	epc_event_serialize_internal(ev, out);
	epc_object_append_oid(out, DM_EV_REASON, &dm_ev->reason->oid);
	epc_object_append_string(out, DM_EV_WHO, dm_ev->who_made);
	epc_object_append_string(out, DM_EV_ACTION, dm_ev->action);
	epc_object_append_object(out, DM_EV_ACTION_DATA, dm_ev->action_data);
}
TEST_ENTRY_POINT(dm_event_serialize);

static struct epc_event_type decision_made_event_type = {
	.name = DECISION_MADE_EVENT_ID,
	.default_ops = {
		.release = dm_event_release,
		.serialize = dm_event_serialize,
	},
	.allocate_event = allocate_dm_event,
	.deserialize_event = deserialize_dm_event,
	.node = LIST_HEAD_INIT(decision_made_event_type.node),
};

EPC_EVENT_TYPE_REGISTER(decision_made_event_type, decision_made_et)
