/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "system_booted_event.h"

static int allocate_sb_event(struct epc_event_type *type,
							 void *data, struct epc_event **ev)
{
	struct system_booted_event *sb_ev;
	struct sb_event_data *sb_ev_data = data;
	int ret;

	sb_ev = malloc(sizeof(*sb_ev));
	if (!sb_ev)
		return -ENOMEM;
	memset(sb_ev, 0, sizeof(*sb_ev));

	sb_ev->boot_id = sb_ev_data->boot_id;
	sb_ev->prev_boot_event = sb_ev_data->prev_boot_event;

	ret = epc_event_init_internal(type, &sb_ev->event);
	if (ret)
		goto free_sb_ev;

	*ev = &sb_ev->event;

	return 0;
free_sb_ev:
	free(sb_ev);
	return ret;
}

static int deserialize_sb_event(struct epc_event_type *type,
								struct epc_object *data, struct epc_event **ev)
{
	int ret = -EINVAL;
	struct sb_event_data sb_ev_data;
	struct epc_object *obj;

	memset(&sb_ev_data, 0, sizeof(sb_ev_data));

	list_for_each_entry(obj, &data->val.children, node) {
		if ((obj->type == TYPE_UUID) &&
				   (strcmp(SB_EV_BID, obj->key) == 0)) {

			sb_ev_data.boot_id = obj->val.uuid;

		} else if ((obj->type == TYPE_OID) &&
				   (strcmp(SB_EV_PREV_BEV, obj->key) == 0)) {

			sb_ev_data.prev_boot_event = obj->val.oid;

		}
	}

	ret = allocate_sb_event(type, &sb_ev_data, ev);
	if (ret < 0)
		goto finish;

	ret = epc_event_deserialize_internal(data, type, *ev);
	if (ret < 0) {
		struct system_booted_event *sb_ev =
			to_system_booted_event(*ev);
		free(sb_ev);
		goto finish;
	}

	ret = 0;
finish:
	return ret;
}
TEST_ENTRY_POINT(deserialize_sb_event);

static void sb_event_release(struct epc_event *ev)
{
	struct system_booted_event *sb_ev =
		to_system_booted_event(ev);

	epc_event_cleanup_internal(&sb_ev->event);
	free(sb_ev);
}

static void sb_event_serialize(struct epc_event *ev, struct epc_object *out)
{
	struct system_booted_event *sb_ev =
		to_system_booted_event(ev);

	epc_event_serialize_internal(ev, out);
	epc_object_append_uuid(out, SB_EV_BID, &sb_ev->boot_id);
	epc_object_append_oid(out, SB_EV_PREV_BEV, &sb_ev->prev_boot_event);
}
TEST_ENTRY_POINT(sb_event_serialize);

static struct epc_event_type system_booted_event_type = {
	.name = SYSTEM_BOOTED_EVENT_ID,
	.default_ops = {
		.release = sb_event_release,
		.serialize = sb_event_serialize,
	},
	.allocate_event = allocate_sb_event,
	.deserialize_event = deserialize_sb_event,
	.node = LIST_HEAD_INIT(system_booted_event_type.node),
};

EPC_EVENT_TYPE_REGISTER(system_booted_event_type, system_booted_et)
