/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "unit_control_event.h"

static int allocate_uc_event(struct epc_event_type *type,
							 void *data, struct epc_event **ev)
{
	struct unit_control_event *uc_ev;
	struct unit_control_event_data *uc_ev_data = data;
	int ret;

	uc_ev = calloc(1, sizeof(*uc_ev));
	if (!uc_ev)
		return -ENOMEM;

	ret = epc_event_init_internal(type, &uc_ev->event);
	if (ret)
		goto cleanup;

	uc_ev->event_time = uc_ev_data->event_time;
	uc_ev->method = uc_ev_data->method;
	uc_ev->unit = uc_ev_data->unit;
	uc_ev->m = uc_ev_data->m;
	uc_ev->jobs_cnt = 0;
	uc_ev->err_code = 0;

	*ev = &uc_ev->event;
	return 0;
cleanup:
	free(uc_ev);

	return ret;
}

static int deserialize_uc_event(struct epc_event_type *type,
								struct epc_object *data, struct epc_event **ev)
{
	int ret = -EINVAL;
	struct unit_control_event_data uc_ev_data;
	struct epc_object *obj;
	memset(&uc_ev_data, 0, sizeof(uc_ev_data));

	list_for_each_entry(obj, &data->val.children, node) {
		switch (obj->type) {
		case TYPE_TIME_T:
			if (!strcmp(UC_TIME, obj->key))
				uc_ev_data.event_time = obj->val.time;
			break;
		case TYPE_STRING:
			if (!strcmp(UC_METHOD, obj->key)) {
				uc_ev_data.method = obj->val.s;
			} else if (!strcmp(UC_UNIT, obj->key)) {
				uc_ev_data.unit = obj->val.s;
			}
			break;
		}
	}

	ret = allocate_uc_event(type, &uc_ev_data, ev);
	if (ret < 0)
		goto finish;

	ret = epc_event_deserialize_internal(data, type, *ev);
	if (ret < 0) {
		struct unit_control_event *uc_ev =
			to_unit_control_event(*ev);
		free(uc_ev);
		goto finish;
	}
	ret = 0;
finish:
	return ret;
}

static void uc_event_release(struct epc_event *ev)
{
	struct unit_control_event *uc_ev =
		to_unit_control_event(ev);

	if (uc_ev->method)
		free(uc_ev->method);
	if (uc_ev->unit)
		free(uc_ev->unit);

	epc_event_cleanup_internal(&uc_ev->event);
	free(uc_ev);
}

static void uc_event_serialize(struct epc_event *ev, struct epc_object *out)
{
	struct unit_control_event *uc_ev =
		to_unit_control_event(ev);
	epc_event_serialize_internal(ev, out);
	epc_object_append_string(out, UC_METHOD, uc_ev->method);
	epc_object_append_time_t(out, UC_TIME, uc_ev->event_time);
}

static struct epc_event_type unit_control_event_type = {
	.name = UC_EVENT_ID,
	.default_ops = {
		.release = uc_event_release,
		.serialize = uc_event_serialize,
	},
	.allocate_event = allocate_uc_event,
	.deserialize_event = deserialize_uc_event,
	.node = LIST_HEAD_INIT(unit_control_event_type.node),
};

EPC_EVENT_TYPE_REGISTER(unit_control_event_type, unit_control_et)
