/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "vconf_key_changed_event.h"

static int allocate_vkc_event(struct epc_event_type *type,
							 void *data, struct epc_event **ev)
{
	struct vconf_key_changed_event *vkc_ev;
	struct vconf_key_changed_event_data *vkc_ev_data = data;
	int ret;

	vkc_ev = calloc(1, sizeof(*vkc_ev));
	if (!vkc_ev)
		return -ENOMEM;

	ret = epc_event_init_internal(type, &vkc_ev->event);
	if (ret)
		goto cleanup;

	vkc_ev->event_time = vkc_ev_data->event_time;
	vkc_ev->key_name = vkc_ev_data->key_name;
	vkc_ev->oldval = vkc_ev_data->oldval;
	vkc_ev->newval = vkc_ev_data->newval;

	*ev = &vkc_ev->event;
	return 0;
cleanup:
	free(vkc_ev);

	return ret;
}

static int deserialize_vkc_event(struct epc_event_type *type,
								struct epc_object *data, struct epc_event **ev)
{
	int ret = -EINVAL;
	struct vconf_key_changed_event_data vkc_ev_data;
	struct epc_object *obj;
	memset(&vkc_ev_data, 0, sizeof(vkc_ev_data));

	list_for_each_entry(obj, &data->val.children, node) {
		const bool oldval_match = !strcmp(VKC_EV_OLDVAL, obj->key);
		const bool newval_match = !strcmp(VKC_EV_NEWVAL, obj->key);

		switch (obj->type) {
		case TYPE_TIME_T:
			if (!strcmp(VKC_EV_TIME, obj->key))
				vkc_ev_data.event_time = obj->val.time;
			break;
		case TYPE_BOOL:
			if (oldval_match) {
				vkc_ev_data.oldval.value.b = obj->val.b;
				vkc_ev_data.oldval.type = VKC_BOOL;
			} else if (newval_match) {
				vkc_ev_data.newval.value.b = obj->val.b;
				vkc_ev_data.newval.type = VKC_BOOL;
			}
			break;
		case TYPE_INT:
			if (oldval_match) {
				vkc_ev_data.oldval.value.i = obj->val.i;
				vkc_ev_data.oldval.type = VKC_INT;
			} else if (newval_match) {
				vkc_ev_data.newval.value.i = obj->val.i;
				vkc_ev_data.newval.type = VKC_INT;
			}
			break;
		case TYPE_DOUBLE:
			if (oldval_match) {
				vkc_ev_data.oldval.value.d = obj->val.d;
				vkc_ev_data.oldval.type = VKC_DOUBLE;
			} else if (newval_match) {
				vkc_ev_data.newval.value.d = obj->val.d;
				vkc_ev_data.newval.type = VKC_DOUBLE;
			}
			break;
		case TYPE_STRING:
			if (!strcmp(VKC_EV_NAME, obj->key)) {
				vkc_ev_data.key_name = obj->val.s;
			} else if (oldval_match) {
				vkc_ev_data.oldval.value.s = obj->val.s;
				vkc_ev_data.oldval.type = VKC_STRING;
			} else if (newval_match) {
				vkc_ev_data.newval.value.s = obj->val.s;
				vkc_ev_data.newval.type = VKC_STRING;
			}
			break;
		}
	}

	ret = allocate_vkc_event(type, &vkc_ev_data, ev);
	if (ret < 0)
		goto finish;

	ret = epc_event_deserialize_internal(data, type, *ev);
	if (ret < 0) {
		struct vconf_key_changed_event *vkc_ev =
			to_vconf_key_changed_event(*ev);
		free(vkc_ev);
		goto finish;
	}
	ret = 0;
finish:
	return ret;
}

static void vkc_event_release(struct epc_event *ev)
{
	struct vconf_key_changed_event *vkc_ev =
		to_vconf_key_changed_event(ev);

	if (vkc_ev->key_name)
		free(vkc_ev->key_name);
	if (vkc_ev->oldval.type == VKC_STRING)
		free(vkc_ev->oldval.value.s);
	if (vkc_ev->newval.type == VKC_STRING)
		free(vkc_ev->newval.value.s);

	epc_event_cleanup_internal(&vkc_ev->event);
	free(vkc_ev);
}

static void vkc_event_serialize(struct epc_event *ev, struct epc_object *out)
{
	struct vconf_key_changed_event *vkc_ev =
		to_vconf_key_changed_event(ev);
	epc_event_serialize_internal(ev, out);
	epc_object_append_string(out, VKC_EV_NAME, vkc_ev->key_name);

	switch (vkc_ev->oldval.type) {
	case VKC_BOOL:
		epc_object_append_bool(out, VKC_EV_OLDVAL, vkc_ev->oldval.value.b);
		break;
	case VKC_INT:
		epc_object_append_int(out, VKC_EV_OLDVAL, vkc_ev->oldval.value.i);
		break;
	case VKC_DOUBLE:
		epc_object_append_double(out, VKC_EV_OLDVAL, vkc_ev->oldval.value.d);
		break;
	case VKC_STRING:
		epc_object_append_string(out, VKC_EV_OLDVAL, vkc_ev->oldval.value.s);
		break;
	}

	switch (vkc_ev->newval.type) {
	case VKC_BOOL:
		epc_object_append_bool(out, VKC_EV_NEWVAL, vkc_ev->newval.value.b);
		break;
	case VKC_INT:
		epc_object_append_int(out, VKC_EV_NEWVAL, vkc_ev->newval.value.i);
		break;
	case VKC_DOUBLE:
		epc_object_append_double(out, VKC_EV_NEWVAL, vkc_ev->newval.value.d);
		break;
	case VKC_STRING:
		epc_object_append_string(out, VKC_EV_NEWVAL, vkc_ev->newval.value.s);
		break;
	}
	epc_object_append_time_t(out, VKC_EV_TIME, vkc_ev->event_time);
}

static struct epc_event_type vconf_key_changed_event_type = {
	.name = VCONF_KEY_CHANGED_EVENT_ID,
	.default_ops = {
		.release = vkc_event_release,
		.serialize = vkc_event_serialize,
	},
	.allocate_event = allocate_vkc_event,
	.deserialize_event = deserialize_vkc_event,
	.node = LIST_HEAD_INIT(vconf_key_changed_event_type.node),
};

EPC_EVENT_TYPE_REGISTER(vconf_key_changed_event_type, vconf_key_changed_et)
