/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>
#include <gio/gio.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "unit_control.h"

#define UNIT_CONTROL_NAME "org.tizen.activationd"
#define UNIT_CONTROL_INTERFACE "org.tizen.activationd"
#define UNIT_CONTROL_OBJ_PATH "/org/tizen/activationd"

#define SYSTEMD_SERVICE "org.freedesktop.systemd1"
#define SYSTEMD_INTERFACE "org.freedesktop.systemd1.Manager"
#define SYSTEMD_OBJ_PATH "/org/freedesktop/systemd1"

enum {
	CALL_TYPE_ACTD,
	CALL_TYPE_SYSTEMD,
};

struct generic_user_data {
	GDBusConnection *bus;
	actd_unit_cb cb;
	void *user_data;
	int call_type;
};

static gint timeout2glib(int timeout_ms)
{
	/* see glib documentation for explanation */
	return timeout_ms >= 0 ? timeout_ms : (timeout_ms == -2 ? G_MAXINT : -1);
}

static void generic_callback(GObject *source_object, GAsyncResult *res, gpointer user_data)
{
	struct generic_user_data *data = user_data;
	GVariant *var;
	GError *err = NULL;
	int status = UNIT_CONTROL_ERROR;
	const char *result;

	if (!data)
		return;

	var = g_dbus_connection_call_finish(data->bus, res, &err);

	if (err && err->code == G_IO_ERROR_TIMED_OUT) {
		status = UNIT_CONTROL_TIMEOUT;
	} else {
		switch (data->call_type) {
		case CALL_TYPE_SYSTEMD:
			if (var)
				status = UNIT_CONTROL_OK;
			else
				status = UNIT_CONTROL_ERROR;

			break;
		case CALL_TYPE_ACTD:
			if (!var)
				status = UNIT_CONTROL_ERROR;
			else {
				g_variant_get(var, "(s)", &result);
				if (g_strcmp0(result, "ok") == 0)
					status = UNIT_CONTROL_OK;
				else
					status = UNIT_CONTROL_ERROR;
			}

			break;
		}
	}

	data->cb(status, data->user_data);

	if (err)
		g_error_free(err);

	free(data);
}

static const char *translate_to_systemd(const char *method)
{
	if (strcmp(method, "Start") == 0)
		return "StartUnit";
	if (strcmp(method, "Stop") == 0)
		return "StopUnit";
	if (strcmp(method, "Restart") == 0)
		return "RestartUnit";

	return NULL;
}

static int call_uc(GBusType bus_type, const char *method, const char *unit, int timeout_ms)
{
	GVariant *msg = NULL;
	int ret = UNIT_CONTROL_ERROR;

	GDBusConnection *bus = g_bus_get_sync(bus_type, NULL, NULL);
	if (!bus)
		return ret;

	GError *err = NULL;

	if (bus_type == G_BUS_TYPE_SESSION) {
		msg = g_dbus_connection_call_sync(bus,
				SYSTEMD_SERVICE,
				SYSTEMD_OBJ_PATH,
				SYSTEMD_INTERFACE,
				translate_to_systemd(method),
				g_variant_new("(ss)", unit, "replace"),
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				timeout2glib(timeout_ms),
				NULL,
				&err);
		if (!msg)
			goto out;
	} else {
		/* we assume that activationd runs on system bus and can be used as a proxy */
		char *result = NULL;

		msg = g_dbus_connection_call_sync(bus,
				UNIT_CONTROL_NAME,
				UNIT_CONTROL_OBJ_PATH,
				UNIT_CONTROL_INTERFACE,
				method,
				g_variant_new("(s)", unit),
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				timeout2glib(timeout_ms),
				NULL,
				&err);
		if (!msg)
			goto out;

		g_variant_get(msg, "(s)", &result);
		bool is_ok = g_strcmp0(result, "ok") == 0;
		g_free(result);

		if (!is_ok)
			goto out;
	}

	ret = UNIT_CONTROL_OK;

out:
	if (err) {
		if (err->code == G_IO_ERROR_TIMED_OUT)
			ret = UNIT_CONTROL_TIMEOUT;
		g_error_free(err);
	}

	return ret;
}

static int call_uc_async(GBusType bus_type, const char *method, const char *unit, actd_unit_cb cb, void *user_data, int timeout_ms)
{
	GDBusConnection *bus;
	struct generic_user_data *data = user_data;
	int ret = UNIT_CONTROL_ERROR;

	bus = g_bus_get_sync(bus_type, NULL, NULL);
	if (!bus)
		return ret;

	data = malloc(sizeof(*data));
	if (!data)
		return ret;

	data->user_data = user_data;
	data->cb = cb;
	data->bus = bus;

	if (bus_type == G_BUS_TYPE_SESSION) {
		data->call_type = CALL_TYPE_SYSTEMD;
		g_dbus_connection_call(bus,
				SYSTEMD_SERVICE,
				SYSTEMD_OBJ_PATH,
				SYSTEMD_INTERFACE,
				translate_to_systemd(method),
				g_variant_new("(ss)", unit, "replace"),
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				timeout2glib(timeout_ms),
				NULL,
				(GAsyncReadyCallback) generic_callback,
				data);
	} else {
		data->call_type = CALL_TYPE_ACTD;
		g_dbus_connection_call(bus,
				UNIT_CONTROL_NAME,
				UNIT_CONTROL_OBJ_PATH,
				UNIT_CONTROL_INTERFACE,
				method,
				g_variant_new("(s)", unit),
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				timeout2glib(timeout_ms),
				NULL,
				(GAsyncReadyCallback) generic_callback,
				data);
	}

	return UNIT_CONTROL_REQUEST_SENT;
}

int actd_start_unit(BusType bus_type, const char *unit, int timeout_ms)
{
	return call_uc((GBusType)bus_type, "Start", unit, timeout_ms);
}

int actd_stop_unit(BusType bus_type, const char *unit, int timeout_ms)
{
	return call_uc((GBusType)bus_type, "Stop", unit, timeout_ms);
}

int actd_restart_unit(BusType bus_type, const char *unit, int timeout_ms)
{
	return call_uc((GBusType)bus_type, "Restart", unit, timeout_ms);
}

int actd_start_unit_async(BusType bus_type, const char *unit, actd_unit_cb cb, void *user_data, int timeout_ms)
{
	return call_uc_async((GBusType)bus_type, "Start", unit, cb, user_data, timeout_ms);
}

int actd_stop_unit_async(BusType bus_type, const char *unit, actd_unit_cb cb, void *user_data, int timeout_ms)
{
	return call_uc_async((GBusType)bus_type, "Stop", unit, cb, user_data, timeout_ms);
}

int actd_restart_unit_async(BusType bus_type, const char *unit, actd_unit_cb cb, void *user_data, int timeout_ms)
{
	return call_uc_async((GBusType)bus_type, "Restart", unit, cb, user_data, timeout_ms);
}
