/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "log.h"
#include "module.h"
#include "common.h"
#include "list.h"
#include "json-config.h"
#include "event_processor.h"
#include "systemd_dbus.h"
#include "dbus_signal_event.h"

#include <glib.h>
#include <gio/gio.h>
#include <libsyscommon/dbus-system.h>

struct dbus_signal {
	char *event_id;
	char *sender;
	char *interface;
	char *member;
	char *path;
	char *path_namespace;
	char *destination;
	struct list_head node;
};

struct dbus_listener {
	struct epc_module module;
	struct list_head signals;
};

#define to_dbus_listener(MOD)						\
	container_of(MOD, struct dbus_listener, module)

static int append_param(struct epc_object *params, enum epc_object_type type, int pos, const void *val)
{
	int r;
	char key[64];

	memset (key, 0, sizeof key);

	r = snprintf(key, sizeof key, "arg%d", pos);
	if (r < 0)
		return r;

	if (type == TYPE_STRING)
		r = epc_object_append_string(params, key, val);
	else
		r = epc_object_append_new(params, key, type, val);

	if (r < 0)
		return r;

	return 0;
}

static int parse_message_args(GVariant *m, struct epc_object *params)
{
	int r;
	int pos = 0;
	GVariantIter iter;
	GVariant *child;

	g_variant_iter_init(&iter, m);
	while (1) {
		child = g_variant_iter_next_value(&iter);
		if (!child)
			break;

		pos++;

		if (g_variant_is_of_type(child, G_VARIANT_TYPE_STRING)){
			const char *v;

			v = g_variant_get_string(child, NULL);
			r = append_param(params, TYPE_STRING, pos, v);
		} else if (g_variant_is_of_type(child, G_VARIANT_TYPE_BOOLEAN)) {
			gboolean v;

			v = g_variant_get_boolean(child);
			r = append_param(params, TYPE_BOOL, pos, &v);
		} else if (g_variant_is_of_type(child, G_VARIANT_TYPE_UINT64)) {
			guint64 v;

			v = g_variant_get_uint64(child);
			r = append_param(params, TYPE_INT, pos, &v);
		} else if (g_variant_is_of_type(child, G_VARIANT_TYPE_INT64)) {
			gint64 v;

			v = g_variant_get_int64(child);
			r = append_param(params, TYPE_INT, pos, &v);
		} else if (g_variant_is_of_type(child, G_VARIANT_TYPE_UINT32)) {
			guint32 v;

			v = g_variant_get_uint32(child);
			r = append_param(params, TYPE_INT, pos, &v);
		} else if (g_variant_is_of_type(child, G_VARIANT_TYPE_INT32)) {
			gint32 v;

			v = g_variant_get_int32(child);
			r = append_param(params, TYPE_INT, pos, &v);
		} else if (g_variant_is_of_type(child, G_VARIANT_TYPE_DOUBLE)) {
			gdouble v;

			v = g_variant_get_double(child);
			r = append_param(params, TYPE_DOUBLE, pos, &v);
		} else {
			r = 0;
		}

		if (r < 0)
			return r;

	}
	return 0;
}


static void on_dbus_signal_match(GDBusConnection *connection,
		const gchar *sender_name,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *signal_name,
		GVariant *parameters,
		gpointer userdata)
{
	struct dbus_signal *s = (struct dbus_signal *)userdata;

	struct dbus_signal_event_data ds_ev_data = {};
	struct epc_event *ev;
	struct timespec ts;
	int ret = 0;

	if (clock_gettime(CLOCK_MONOTONIC, &ts) < 0) {
		log_error_errno(errno, "Unable to get timestamp : %m");
		return;
	}

	ds_ev_data.event_time = ts.tv_sec;

#define COPY_COND(x) do { \
		if (s->x) { \
			ds_ev_data.x = strdup(s->x); \
			if (ds_ev_data.x == NULL) \
				goto finish; \
		} \
	} while (0)

	COPY_COND(event_id);
	COPY_COND(path_namespace);
	COPY_COND(destination);

#undef COPY_COND

	// copy actual signal data
	// While it's possible to copy some of the values from userdata, which was
	// used during signal registration but at least path must be copied from
	// function parameters
	ds_ev_data.sender = strdup(sender_name);
	ds_ev_data.path = strdup(object_path);
	ds_ev_data.interface = strdup(interface_name);
	ds_ev_data.member = strdup(signal_name);
	if (!ds_ev_data.sender || !ds_ev_data.path || !ds_ev_data.interface || !ds_ev_data.member)
		// XXX handle individual errors
		goto finish;

	if (epc_object_new(&ds_ev_data.match_params) < 0)
		goto finish;

	ret = parse_message_args(parameters, ds_ev_data.match_params);
	if (ret) {
		log_error_errno(ret, "Unable to parse signal parameters: %m.");
		goto finish;
	}

	if(parameters) {
		g_variant_ref(parameters);
		ds_ev_data.signal_params = parameters;
	}

	ret = epc_event_create(DBUS_SIGNAL_EVENT_ID, &ds_ev_data, &ev);
	if (ret) {
		log_error_errno(ret, "Unable to allocate an event: %m.");
		goto finish;
	}

	ret = event_processor_report_event(ev);
	epc_event_unref(ev);
	if (ret)
		log_error_errno(ret, "Unable to report event: %m");

	return;

finish:
	free(ds_ev_data.event_id);
	free(ds_ev_data.sender);
	free(ds_ev_data.interface);
	free(ds_ev_data.member);
	free(ds_ev_data.path);
	free(ds_ev_data.path_namespace);
	free(ds_ev_data.destination);
	if (ds_ev_data.match_params)
		epc_object_destroy(ds_ev_data.match_params);
}

static void signal_cleanup(struct dbus_signal *s)
{
	free(s->event_id);
	free(s->sender);
	free(s->interface);
	free(s->member);
	free(s->path);
	free(s->path_namespace);
	free(s->destination);
	free(s);
}


static void cleanup(struct dbus_listener *l)
{
	struct dbus_signal *s, *next;

	list_for_each_entry_safe(s, next, &l->signals, node) {
		list_del(&s->node);
		signal_cleanup(s);
	}
}

static int add_signal(struct dbus_listener *l, json_object *root, dbus_handle_h bus)
{
	int ret;
	struct dbus_signal *s;

	s = calloc(1, sizeof(*s));
	if (!s) {
		log_error("Could not allocate dbus_signal object");
		return -ENOMEM;
	}

	ret = get_config_field(root, "id", &s->event_id, json_type_string);
	if (ret < 0)
		goto out;

	log_debug("Adding signal: %s", s->event_id);

	(void) get_config_field(root, "sender", &s->sender, json_type_string);

	ret = get_config_field(root, "interface", &s->interface, json_type_string);
	if (ret < 0)
		goto out;

	ret = get_config_field(root, "member", &s->member, json_type_string);
	if (ret < 0)
		goto out;

	(void) get_config_field(root, "path", &s->path, json_type_string);
	(void) get_config_field(root, "path_namespace", &s->path_namespace, json_type_string);

	if (!s->path && !s->path_namespace)
		goto out;

	(void) get_config_field(root, "destination", &s->destination, json_type_string);

	ret = subscribe_dbus_signal(bus, s->path, s->interface, s->member, on_dbus_signal_match, s, NULL);
	if (ret < 0) {
		log_error("Failed to subscribe to dbus signals\n");
		goto out;
	}

	list_add_tail(&s->node, &l->signals);
	return 0;

out:
	signal_cleanup(s);
	return ret;
}


static int dbus_listener_init(struct epc_module *module,
								 struct epc_config *config,
								 sd_event* loop)
{
	struct dbus_listener *listener = to_dbus_listener(module);
	int ret = 0, len, i;
	json_object *arr, *obj;
	dbus_handle_h bus;

	INIT_LIST_HEAD(&listener->signals);

	if (config == NULL)
		return 0;

	bus = dbus_handle_get_connection(G_BUS_TYPE_SYSTEM, 0);
	if (!bus) {
		log_error_errno(ret, "Failed to acquire the default system bus connection: %m");
		return -EINVAL;
	}

	if (!json_object_object_get_ex(config->root, "signals", &arr)) {
		log_error("There is no 'signals' array");
		return -EINVAL;
	}

	if (!json_object_is_type(arr, json_type_array)) {
		log_error("Config value is not an array");
		return -EINVAL;
	}

	len = json_object_array_length(arr);
	for (i = 0; i < len; ++i) {
		obj = json_object_array_get_idx(arr, i);
		ret = add_signal(listener, obj, bus);
		if (ret < 0) {
			log_error("Could not add signal");
			goto cleanup;
		}
	}

	return 0;

cleanup:
	cleanup(listener);
	return ret;
}

static void dbus_listener_cleanup(struct epc_module *module)
{
	struct dbus_listener *listener = to_dbus_listener(module);
	cleanup(listener);
}

struct dbus_listener dbus_listener = {
	.module = {
		.name = "dbus_listener",
		.type = EPC_MODULE_TYPE_LISTENER,

		.init = dbus_listener_init,
		.cleanup = dbus_listener_cleanup,
		.node = LIST_HEAD_INIT(dbus_listener.module.node),
	},
};

EPC_MODULE_REGISTER(&dbus_listener.module)
