/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "log.h"
#include "module.h"
#include "common.h"
#include "event_processor.h"
#include "systemd_dbus.h"
#include "unit_control_event.h"
#include <libsyscommon/dbus-system.h>

#define ACTD_DBUS_API_INTERFACE_NAME "org.tizen.activationd"
#define ACTD_DBUS_API_SERVICE_NAME "org.tizen.activationd"
#define ACTD_DBUS_API_OBJECT_PATH "/org/tizen/activationd"

#define STRING_VARIANT	((const GVariantType *) "(s)")

static GVariant *method_generic_callback(GVariant *param, GDBusMethodInvocation *invocation,
		void *userdata, const char *method)
{
	int r;
	const char *s;
	struct unit_control_event_data uc_ev_data = {};
	struct epc_event *ev;
	struct timespec ts;

	if (strcmp(method, "start") &&
			strcmp(method, "stop") &&
			strcmp(method, "restart")) {
		r = -EINVAL;
		goto fail;
	}

	if (!g_variant_is_of_type(param, STRING_VARIANT)) {
		log_error("Invalid parameter type passed to %s method.  "
			  "Expected: (s), got: %s\n", method, g_variant_get_type_string(param));
		r = -EINVAL;
		goto fail;
	}

	g_variant_get(param, "(s)", &s);

	if (clock_gettime(CLOCK_MONOTONIC, &ts) < 0) {
		log_error_errno(errno, "Unable to get timestamp : %m");
		r = -errno;
		goto fail;
	}

	uc_ev_data.event_time = ts.tv_sec;

	uc_ev_data.unit = strdup(s);
	if (uc_ev_data.unit == NULL) {
		r = -ENOMEM;
		goto fail;
	}

	uc_ev_data.method = strdup(method);
	if (uc_ev_data.method == NULL) {
		r = -ENOMEM;
		goto fail;
	}

	/*we'll use it later for the response */
	/* XXX ref(invocation) */
	uc_ev_data.m = invocation;

	r = epc_event_create(UC_EVENT_ID, &uc_ev_data, &ev);
	if (r) {
		log_error_errno(r, "Unable to allocate an event: %m.");
		r = -errno;
		goto fail;
	}

	r = event_processor_report_event(ev);
	epc_event_unref(ev);
	if (r) {
		log_error_errno(r, "Unable to report event: %m");
		r = -EINVAL;
		goto fail;
	}

	return NULL;

fail:

	free(uc_ev_data.method);
	free(uc_ev_data.unit);
	return g_variant_new("(s)", "error");
}

static GVariant *method_start(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return method_generic_callback(param, invocation, user_data, "start");
}

static GVariant *method_stop(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return method_generic_callback(param, invocation, user_data, "stop");
}

static GVariant *method_restart(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return method_generic_callback(param, invocation, user_data, "restart");
}

static const dbus_method_s unit_control_api_vtable[] = {
	{"Start", "s", "s", method_start},
	{"Stop", "s", "s", method_stop},
	{"Restart", "s", "s", method_restart},
};

static const dbus_interface_u unit_control_api_interface = {
	.oh = NULL,
	.name = ACTD_DBUS_API_INTERFACE_NAME,
	.methods = unit_control_api_vtable,
	.nr_methods = ARRAY_SIZE(unit_control_api_vtable),
};

struct unit_control_api {
	struct epc_module module;
	dbus_handle_h bus;
};

#define to_unit_control_api(MOD)						\
	container_of(MOD, struct unit_control_api, module)

static void cleanup_listener(struct unit_control_api *l)
{
}

static void actd_dbus_name_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data)
{
	log_debug("Name acquired");
}


static int unit_control_api_init(struct epc_module *module,
								 struct epc_config *config,
								 sd_event* loop)
{
	struct unit_control_api *listener = to_unit_control_api(module);
	int ret = 0;

	listener->bus = dbus_handle_get_connection(G_BUS_TYPE_SYSTEM, 0);
	if (!listener->bus) {
		log_error_errno(ret, "Failed to acquire the default system bus connection: %m");
		return -EINVAL;
	}

	ret = dbus_handle_register_dbus_object(listener->bus,
			ACTD_DBUS_API_OBJECT_PATH,
			&unit_control_api_interface);
	if (ret < 0) {
		fprintf(stderr, "Failed to register api calls: %s\n", strerror(-ret));
		goto cleanup;
	}

	/* Take a well-known service name so that clients can find us */
	ret = dbus_handle_request_bus_name(listener->bus, ACTD_DBUS_API_SERVICE_NAME, actd_dbus_name_acquired, NULL);
	if (ret < 0) {
		fprintf(stderr, "Failed to acquire service name: %s\n", strerror(-ret));
		goto cleanup;
	}

	return 0;

cleanup:
	cleanup_listener(listener);
	return ret;
}

static void unit_control_api_cleanup(struct epc_module *module)
{
	struct unit_control_api *listener = to_unit_control_api(module);
	cleanup_listener(listener);
}

struct unit_control_api unit_control_api = {
	.module = {
		.name = "unit_control_api",
		.type = EPC_MODULE_TYPE_LISTENER,

		.init = unit_control_api_init,
		.cleanup = unit_control_api_cleanup,
		.node = LIST_HEAD_INIT(unit_control_api.module.node),
	},
};

EPC_MODULE_REGISTER(&unit_control_api.module)
