/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <vconf.h>
#include <glib.h>
#include <time.h>

#include "log.h"
#include "module.h"
#include "common.h"
#include "list.h"
#include "json-config.h"
#include "event_processor.h"
#include "vconf_key_changed_event.h"

struct vconf_kv {
	char *name;
	struct vkc_value val;
	struct list_head node;
};

struct vconf_listener {
	struct epc_module module;
	struct list_head keys;
};

#define to_vconf_listener(MOD)						\
	container_of(MOD, struct vconf_listener, module)

static struct vconf_kv *find_vconf_kv(struct vconf_listener *l, const char *name)
{
	struct vconf_kv *k = NULL;

	list_for_each_entry(k, &l->keys, node) {
		if (!strcmp(k->name, name))
			return k;
	}
	return NULL;
}

static int copy_vkc_value(struct vkc_value *from, struct vkc_value *to)
{
	to->type = from->type;
	switch (from->type){
	case VKC_BOOL:
		to->value.b = from->value.b;
		break;
	case VKC_INT:
		to->value.i = from->value.i;
		break;
	case VKC_DOUBLE:
		to->value.d = from->value.d;
		break;
	case VKC_STRING:
		to->value.s = strdup(from->value.s);
		if (!to->value.s)
			return -ENOMEM;
		break;
	}
	return 0;
}

static int set_object_value(struct vconf_kv *k, keynode_t *node)
{
	assert (k);
	assert (node);
	int ret = 0;

	switch(vconf_keynode_get_type(node)) {
	case VCONF_TYPE_INT:{
		int value = vconf_keynode_get_int(node);
		k->val.value.i = value;
		k->val.type = VKC_INT;
		}break;
	case VCONF_TYPE_BOOL:{
		int value = vconf_keynode_get_bool(node);
		if (value ==  -1) {
			ret = -1;
			break;
		}
		k->val.value.b = value;
		k->val.type = VKC_BOOL;
		}break;
	case VCONF_TYPE_DOUBLE:{
		double value = vconf_keynode_get_dbl(node);
		k->val.value.d = value;
		k->val.type = VKC_DOUBLE;
		}break;
	case VCONF_TYPE_STRING:{
		char *value = vconf_keynode_get_str(node);
		if (!value) {
			ret = -1;
			break;
		}
		k->val.value.s = strdup(value);
		if (!k->val.value.s)
			return -1;
		k->val.type = VKC_STRING;
		}break;
	default:
		ret = -1;
		break;
	}

	return ret;
}

static void on_vconf_key_changed(keynode_t *key, void *data)
{
	struct vconf_listener *l = (struct vconf_listener *)data;
	char *name = NULL;
	struct vconf_key_changed_event_data vkc_ev_data = {};
	struct epc_event *ev;
	struct timespec ts;
	int ret = 0;

	if (clock_gettime(CLOCK_MONOTONIC, &ts) < 0) {
		log_error_errno(errno, "Unable to get timestamp : %m");
		return;
	}

	name = vconf_keynode_get_name(key);
	if (!name) {
		log_debug("vconf:  key name not found (key=%s)", name);
		return;
	}

	struct vconf_kv *k = find_vconf_kv(l, name);
	if (!k) {
		log_debug("vconf: key-value pair not found (key=%s)", name);
		return;
	}

	vkc_ev_data.event_time = ts.tv_sec;
	vkc_ev_data.key_name = strdup(name);
	if (vkc_ev_data.key_name == NULL)
		goto finish;

	if (copy_vkc_value(&k->val, &vkc_ev_data.oldval) < 0)
		goto finish;

	if (k->val.type == VKC_STRING) {
		free(k->val.value.s);
		k->val.value.s = NULL;
	}

	if (set_object_value(k, key) < 0)
		goto finish;

	if (copy_vkc_value(&k->val, &vkc_ev_data.newval) < 0)
		goto finish;

	ret = epc_event_create(VCONF_KEY_CHANGED_EVENT_ID, &vkc_ev_data, &ev);
	if (ret) {
		log_error_errno(ret, "Unable to allocate an event: %m.");
		goto finish;
	}

	ret = event_processor_report_event(ev);
	epc_event_unref(ev);
	if (ret)
		log_error_errno(ret, "Unable to report event: %m");
	return;
finish:
	free(vkc_ev_data.key_name);
	if (vkc_ev_data.oldval.type == VKC_STRING)
		free(vkc_ev_data.oldval.value.s);
	if (vkc_ev_data.newval.type == VKC_STRING)
		free(vkc_ev_data.newval.value.s);

	return;
}

static int add_key(struct vconf_listener *l, const char *key, keynode_t *node)
{
	int ret = 0;
	struct vconf_kv *k;

	assert (l);
	assert (key);
	//assert (node); /* it's ok to be null */

	k = (struct vconf_kv*)malloc(sizeof(struct vconf_kv));
	if (k == NULL) {
		ret = -ENOMEM;
		goto cleanup;
	}

	k->name = strdup(key);
	if (k->name == NULL) {
		ret = -ENOMEM;
		goto cleanup;
	}

	if (node != NULL && set_object_value(k, node) < 0) {
		ret = -EINVAL;
		goto cleanup;
	}

	list_add_tail(&k->node, &l->keys);
	if (vconf_notify_key_changed(key, on_vconf_key_changed, l) < 0)
		log_error("vconf_notify_key_changed() failed (%s)", key);

	return 0;

cleanup:
	if (k) {
		free(k->name);
		free(k);
	}
	return ret;
}

static void cleanup(struct vconf_listener *l)
{
	struct vconf_kv *k, *next;

	list_for_each_entry_safe(k, next, &l->keys, node) {
		list_del(&k->node);
		free(k->name);
		if (k->val.type == VKC_STRING)
			free(k->val.value.s);
		free(k);
	}
}

static int vconf_listener_init(struct epc_module *module,
								 struct epc_config *config,
								 sd_event* event)
{
	struct vconf_listener *listener = to_vconf_listener(module);
	int ret = 0;
	json_object *arr, *val;
	keylist_t *klist;
	keynode_t *knode;

	INIT_LIST_HEAD(&listener->keys);

	if (!json_object_object_get_ex(config->root, "keys", &arr)) {
		log_error("Config does not contain 'keys' parameter.");
		ret = -EINVAL;
		goto cleanup;
	}

	if (!json_object_is_type(arr, json_type_array)) {
		log_error("Config value is not an array");
		ret = -EINVAL;
		goto cleanup;
	}

	for (size_t i = 0; i < json_object_array_length(arr); i++) {
		knode = NULL;
		val = json_object_array_get_idx(arr, i);
		if (val == NULL) {
			log_error("vconf: no specified key (%zu)\n", i);
			ret = -EINVAL;
			goto cleanup;
		}

		klist = vconf_keylist_new();
		if (!klist) {
			log_error("vconf: could not create keylist for %s\n", json_object_get_string(val));
			continue;
		}

		ret = vconf_get(klist, json_object_get_string(val), VCONF_GET_ALL);
		if (ret < 0) {
			log_debug("vconf: keylist does not contain key %s\n", json_object_get_string(val));
			goto free_list;
		}

		ret = vconf_keylist_lookup(klist, json_object_get_string(val), &knode);
		if (ret <= 0) {
			log_debug("vconf_keylist_lookup failed for %s\n", json_object_get_string(val));
			//goto free_list;
		}

		ret = add_key(listener, json_object_get_string(val), knode);
		if (ret < 0)
			log_error("vconf: could not add key\n");

		free_list:
			vconf_keylist_free(klist);
	}
	return 0;

cleanup:
	cleanup(listener);
	return ret;
}

static void vconf_listener_cleanup(struct epc_module *module)
{
	struct vconf_listener *listener = to_vconf_listener(module);
	cleanup(listener);
}

struct vconf_listener vconf_listener = {
	.module = {
		.name = "vconf_listener",
		.type = EPC_MODULE_TYPE_LISTENER,

		.init = vconf_listener_init,
		.cleanup = vconf_listener_cleanup,
		.node = LIST_HEAD_INIT(vconf_listener.module.node),
	},
};

EPC_MODULE_REGISTER(&vconf_listener.module)
