/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

#include "common.h"

#define APPEND_STR(src, format, ...) ({\
	!*src ? asprintf(src, format, ##__VA_ARGS__) : \
	asprintf_with_free(src, "%s" format, *src, ##__VA_ARGS__); \
})

static int (*at_exit)(void) = NULL;

void epc_set_exit_handler(int (*handler)(void))
{
	at_exit = handler;
}

int epc_call_at_exit(void)
{
	return (at_exit) != NULL ? at_exit() : 0;
}

static const size_t data_size[] = {
	[TYPE_OID] = sizeof(epc_oid_t),
	[TYPE_STRING] = sizeof(char *),
	[TYPE_INT] = sizeof(int),
	[TYPE_TIMESPEC] = sizeof(struct timespec),
	[TYPE_TIME_T] = sizeof(time_t),
	[TYPE_BOOL] = sizeof(bool),
	[TYPE_DOUBLE] = sizeof(double),
	[TYPE_UUID] = sizeof(sd_id128_t),
	[TYPE_OBJECT] = sizeof(struct list_head),
};

static char *epc_object_type_names[] = {
	[TYPE_OID] = "TYPE_OID",
	[TYPE_STRING] = "TYPE_STRING",
	[TYPE_INT] = "TYPE_INT",
	[TYPE_TIMESPEC] = "TYPE_TIMESPEC",
	[TYPE_TIME_T] = "TYPE_TIME_T",
	[TYPE_BOOL] = "TYPE_BOOL",
	[TYPE_DOUBLE] = "TYPE_DOUBLE",
	[TYPE_UUID] = "TYPE_UUID",
	[TYPE_OBJECT] = "TYPE_OBJECT",
};

char* fo_type_to_str(unsigned type)
{
	if (type < ARRAY_SIZE(epc_object_type_names))
		return epc_object_type_names[type];
	else
		return NULL;
}

static inline void copy_obj_val(struct epc_object *obj, const void *val)
{
	memcpy(&obj->val, val, data_size[obj->type]);
}

static void release_epc_object(struct uref *uref)
{
	struct epc_object *obj = container_of(uref, struct epc_object, uref);

	epc_object_destroy(obj);
	free(obj);
}

static inline void init_obj(struct epc_object *obj, int type, const void *val)
{
	assert(val != NULL || type == TYPE_OBJECT);

	obj->type = type;
	obj->key = NULL;
	uref_init(&obj->uref, release_epc_object);
	INIT_LIST_HEAD(&obj->node);

	if (type == TYPE_OBJECT && val == NULL)
		INIT_LIST_HEAD(&obj->val.children);
	else
		copy_obj_val(obj, val);
}

static struct epc_object *new_object_node(int type, const void *val)
{
	struct epc_object *obj;

	assert(type >= 0);

	obj = calloc(1, sizeof(*obj));
	if (!obj)
		return NULL;

	init_obj(obj, type, val);

	return obj;
}

int epc_object_fill_empty(struct epc_object *obj)
{
	return 0;
}

int epc_object_new_flags(struct epc_object **obj, int flags)
{
	assert(obj);

	*obj = new_object_node(TYPE_OBJECT, NULL);
	if (!*obj)
		return -ENOMEM;

	(*obj)->flags = flags;
	return 0;
}

void epc_object_init(struct epc_object *obj)
{
	assert(obj);

	init_obj(obj, TYPE_OBJECT, NULL);
}

void epc_object_destroy(struct epc_object *obj)
{
	struct epc_object *child, *next;

	switch (obj->type) {
	case TYPE_OBJECT:
		list_for_each_entry_safe(child, next, &obj->val.children, node) {
			list_del_init(&child->node);
			epc_object_unref(child);
		}
		break;
	case TYPE_STRING:
		free(obj->val.s);
		break;
	default:
		break;
	};

	free(obj->key);
}

static int object_append(struct epc_object *obj, const char *key, struct epc_object *child)
{
	child->key = strdup(key);
	if (!child->key)
		return -ENOMEM;

	list_add_tail(&child->node, &obj->val.children);
	return 0;
}

int epc_object_append_new(struct epc_object *parent, const char *key,
							 int type, const void *val)
{
	struct epc_object *child;
	int ret;

	assert(parent);
	assert(parent->type == TYPE_OBJECT);
	assert(key);
	assert(val);

	child = new_object_node(type, val);
	if (!child)
		return -ENOMEM;

	ret = object_append(parent, key, child);
	if (ret < 0)
		goto unref_child;

	return 0;

unref_child:
	epc_object_unref(child);
	return ret;
}

int epc_object_append_string(struct epc_object *obj, const char *key, const char *val)
{
	char *val_copy;
	int ret;

	val_copy = strdup(val);
	if (!val_copy)
		return -ENOMEM;

	ret = epc_object_append_new(obj, key, TYPE_STRING, &val_copy);
	if (ret < 0)
		goto free_val_copy;

	return 0;
free_val_copy:
	free(val_copy);
	return ret;
}

int epc_object_append_object(struct epc_object *obj, const char *key,
								struct epc_object *val)
{
	return object_append(obj, key, epc_object_ref(val));
}

static struct epc_object *find_child(struct epc_object *obj, const char *key)
{
	struct epc_object *child;
	size_t len;
	char *dot;

	assert(key);

	dot = strchr(key, '.');
	if (dot)
		len = dot - key;
	else
		len = strlen(key);

	list_for_each_entry(child, &obj->val.children, node) {
		if (strncmp(key, child->key, len))
			continue;

		if (child->key[len] != '\0')
			continue;

		/* is this end of key? */
		if (!dot)
			return child;

		/* Can search we recursively? */
		if (child->type == TYPE_OBJECT)
			return find_child(child, dot + 1);
		else
			break;
	}

	return NULL;
}

int epc_object_get_val(struct epc_object *obj, const char *key,
						  int type, void *val)
{
	struct epc_object *child;

	assert(obj);
	assert(obj->type == TYPE_OBJECT);
	assert(key);

	child = find_child(obj, key);
	if (!child)
		return -ENOENT;

	if (child->type != type)
		return -EFAULT;

	memcpy(val, &child->val, data_size[child->type]);
	return 0;
}

int epc_object_get_object(struct epc_object *obj, const char *key, struct epc_object **val)
{
	struct epc_object *child;

	child = find_child(obj, key);
	if (!child)
		return -ENOENT;

	if (child->type != TYPE_OBJECT)
		return -EFAULT;

	*val = child;
	return 0;
}

static int asprintf_with_free(char **src, const char *format, ...)
{
	char *buf = NULL;
	va_list ap;
	int ret;

	va_start(ap, format);
	ret = vasprintf(&buf, format, ap);
	va_end(ap);

	if (ret < 0)
		return -ENOMEM;

	free(*src);
	*src = buf;

	return 0;
}

int epc_object_to_string(char **str, struct epc_object *parent, int indent)
{
	int ret;

	for (int i = 0; i < indent; ++i) {
		ret = APPEND_STR(str, "\t");
		if (ret < 0)
			goto error;
	}

	ret = APPEND_STR(str, "\"%s\" ", parent->key);
	if (ret < 0)
		goto error;

	switch (parent->type) {
	case TYPE_OID:
		{
			char *oidbuf = epc_oid_to_string_hex(&parent->val.oid);

			ret = APPEND_STR(str, "(%s) : %s\n",
							 fo_type_to_str(TYPE_OID), oidbuf);
			free(oidbuf);
			if (ret < 0)
				goto error;

			break;
		}

	case TYPE_UUID:
		ret = APPEND_STR(str, "(%s) : %s\n",
						 fo_type_to_str(TYPE_UUID), SD_ID128_CONST_STR(parent->val.uuid));
		if (ret < 0)
			goto error;

		break;

	case TYPE_STRING:
		ret = APPEND_STR(str, "(%s) : \"%s\"\n",
						 fo_type_to_str(TYPE_STRING), parent->val.s);
		if (ret < 0)
			goto error;

		break;

	case TYPE_INT:
		ret = APPEND_STR(str, "(%s) : %d\n",
						 fo_type_to_str(TYPE_INT), parent->val.i);
		if (ret < 0)
			goto error;

		break;

	case TYPE_TIME_T:
		ret = APPEND_STR(str, "(%s) : %ld\n",
						 fo_type_to_str(TYPE_TIME_T), parent->val.time);
		if (ret < 0)
			goto error;

		break;

	case TYPE_TIMESPEC:
		ret = APPEND_STR(str, "(%s) : (tv_sec = %ld, tv_nsec = %ld)\n",
						 fo_type_to_str(TYPE_TIMESPEC),
						 parent->val.ts.tv_sec,
						 parent->val.ts.tv_nsec);
		if (ret < 0)
			goto error;

		break;

	case TYPE_OBJECT:
		{
			struct epc_object *child;

			ret = APPEND_STR(str, "(%s) :\n", fo_type_to_str(TYPE_OBJECT));
			if (ret < 0)
				goto error;

			list_for_each_entry(child, &parent->val.children, node) {
				ret = epc_object_to_string(str, child, indent + 1);
				if (ret < 0)
					goto error;
			}

			break;
		}

	default:
		ret = APPEND_STR(str, "(unsupported type %d)\n", parent->type);
		if (ret < 0)
			goto error;

		break;
	}

	return 0;

error:
	free(*str);
	*str = NULL;
	return -ENOMEM;
}

int epc_oid_from_string(epc_oid_t *oid, const char *s)
{
	return sd_id128_from_string(s, &oid->uuid);
}

int epc_oid_equal(const epc_oid_t *oid1, const epc_oid_t *oid2)
{
	if (!oid1 || !oid2)
		return 0;

	return !memcmp(oid1, oid2, sizeof *oid1);
}

char *epc_oid_to_string_hex(const epc_oid_t *oid)
{
	unsigned char *bytes = (unsigned char*)oid;
	unsigned int size = sizeof(*oid);
	char *buf;
	int ret;

	buf = calloc(sizeof(*oid) * 2 + 1, sizeof(char));
	if (!buf)
		return NULL;

	for (unsigned int i = 0; i < size; i++) {
		ret = snprintf(&buf[i * 2], 3, "%02x", bytes[i]);
		if (ret < 0) {
			free(buf);
			return NULL;
		}
	}

	return buf;
}

enum epc_object_type epc_type_from_json(enum json_type type)
{
	switch (type) {
	case json_type_boolean:
		return TYPE_BOOL;
	case json_type_double:
		return TYPE_DOUBLE;
	case json_type_int:
		return TYPE_INT;
	case json_type_object:
		return TYPE_OBJECT;
	case json_type_string:
		return TYPE_STRING;
	case json_type_array:
	case json_type_null:
		return TYPE_UNKNOWN;
	}
	return TYPE_UNKNOWN;
}

bool epc_object_match_pattern(struct epc_object *obj, struct json_object *root)
{
	struct epc_object *node;
	int ret;

	json_object_object_foreach(root, key, value) {
		node = find_child(obj, key);

		if (!node)
			return 0;

		if (node->type != (int)epc_type_from_json(json_object_get_type(value)))
			return 0;

		switch (node->type) {
		case TYPE_STRING:
			if (strcmp(node->val.s, json_object_get_string(value)))
				return 0;
			break;
		case TYPE_DOUBLE:
			if (node->val.d != json_object_get_double(value))
				return 0;
			break;
		case TYPE_INT:
			if (node->val.i != json_object_get_int(value))
				return 0;
			break;
		case TYPE_BOOL:
			if (node->val.b != json_object_get_boolean(value))
				return 0;
			break;
		case TYPE_OBJECT:
			ret = epc_object_match_pattern(node, value);
			if (!ret)
				return 0;
			break;
		default:
			return 0;
		}
	}

	return 1;
}
