/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <string.h>
#include <stdio.h>
#include <json-c/json.h>
#include <dirent.h>
#include <sys/stat.h>

#include "json-config.h"
#include "log.h"

static int is_directory(const char *path)
{
	struct stat st;
	int ret;

	ret = stat(path, &st);
	if (ret == -1) {
		log_error("Unable to stat %s: %m", path);
		return 0;
	}
	return st.st_mode & S_IFDIR;
}

static int check_suffix(const char *path, const char *suffix)
{
	int suffix_len = strlen(suffix);
	int total_len = strlen(path);

	if (total_len <= suffix_len)
		return 0;

	return strcmp(&path[total_len - suffix_len], suffix) == 0;
}

static int remove_last_n_chars(const char *src, size_t n,
							   char *dest, size_t dest_size)
{
	size_t len = strlen(src) - n;

	if (len + 1 > dest_size) {
		log_error("Destination buffer is too small");
		return -EOVERFLOW;
	}

	strncpy(dest, src, len);
	dest[len] = 0;

	return 0;
}

static int config_filter(const struct dirent *d, const char *suffix)
{
	if (d->d_name[0] == '.')
		return 0;

	if (check_suffix(d->d_name, suffix))
		return 1;

	return 0;
}

static int config_dir_filter(const struct dirent *d)
{
	return config_filter(d, ".conf.d");
}

static int config_file_filter(const struct dirent *d)
{
	return config_filter(d, ".conf");
}

static int json_object_object_merge(json_object *obj1, char *key, json_object *obj2_node)
{
	json_object *obj1_node = NULL;
	int ret;

	ret = json_object_object_get_ex(obj1, key, &obj1_node);

	/* if there is no such object in obj1 - add it */
	if (ret == 0)
		goto json_add;

	if (obj1_node == NULL) {
		log_error("Unable to get json object");
		return -ENOMEM;
	}

	/* if both obj1 and obj2 are arrays move contents from
	 * obj1 to obj2 and replace obj1 with obj2 */
	if (json_object_is_type(obj1_node, json_type_array) &&
		json_object_is_type(obj2_node, json_type_array)) {
		size_t obj2_size  = json_object_array_length(obj2_node);
		for (size_t i = 0; i < json_object_array_length(obj1_node); i++) {
			json_object *o = json_object_array_get_idx(obj1_node, i);
			json_object_get(o);
			json_object_array_put_idx(obj2_node, obj2_size++, o);
		}
		/* obj1 will be deleted in next condition (it'll be true) */
	}

	/* if there is object with key in obj1 (obj2), but either obj1 or obj2
	 * is not json_type_object - replace it */
	if (!json_object_is_type(obj1_node, json_type_object) ||
		!json_object_is_type(obj2_node, json_type_object)) {

		json_object_object_del(obj1, key);
		goto json_add;
	}

	/* if obj1 and obj2 is json_object_type */
	json_object_object_foreach(obj2_node, k, v) {
		ret = json_object_object_merge(obj1_node, k, v);
		if (ret)
			return ret;
	}

	return 0;

json_add:
	json_object_get(obj2_node);
	json_object_object_add(obj1, key, obj2_node);
	return 0;
}

static int json_object_merge_with_file(json_object *existing_node, const char *file)
{
	struct json_object *new_node;
	int ret;

	new_node = json_object_from_file(file);
	if (new_node == NULL) {
		log_error("Could not create object from json file %s", file);
		return -EINVAL;
	}

	json_object_object_foreach(new_node, key, val) {
		ret = json_object_object_merge(existing_node, key, val);
		if (ret)
			goto cleanup;
	}

cleanup:
	json_object_put(new_node);
	return ret;
}

static int assemble_config_from_dir(const char *path, json_object *config)
{
	int num;
	struct dirent **list;
	char filename[PATH_MAX];
	int ret;

	/* search for *.conf files for this module */
	num = scandir(path, &list, config_file_filter, alphasort);
	if (num < 0)
		return num;

	while (num--) {
		/* get full path of the config file */
		ret = snprintf(filename, sizeof(filename), "%s/%s", path, list[num]->d_name);
		free(list[num]);
		if (ret >= (int) sizeof(filename)) {
			log_error("Unable to copy string");
			goto error;
		}

		log_debug("Using config file: %s", filename);

		/* merge config into existing json object */
		ret = json_object_merge_with_file(config, filename);
		if (ret < 0)
			goto error;
	}

	free(list);
	return 0;

error:
	while (num--)
		free(list[num]);
	free(list);
	return ret;
}

static int load_multidir_config(const char *path, struct json_object *c)
{
	int num;
	struct dirent **list;
	char dir_path[PATH_MAX], obj_name[PATH_MAX];
	struct json_object *obj_config = NULL;
	int ret;

	/* search for *.conf.d directories */
	num = scandir(path, &list, config_dir_filter, alphasort);
	if (num < 0) {
		/* dir doesn't exist */
		if (errno == ENOENT) {
			log_debug("Config dir doesn't exist %s", path);
			return -ENOENT;
		}

		/* dir exists but we are unable to access it*/
		log_error("Could not access config dir %s", path);
		return -errno;
	}

	while (num--) {
		/* get full path of the config dir */
		ret = snprintf(dir_path, sizeof(dir_path), "%s/%s", path, list[num]->d_name);
		free(list[num]);
		if (ret >= (int) sizeof(dir_path)) {
			log_error("Unable to copy string");
			goto error;
		}

		if (!is_directory(dir_path))
			continue;

		/* get object name out of directory path */
		ret = remove_last_n_chars(basename(dir_path), strlen(".conf.d"),
								  obj_name, sizeof(obj_name));
		if (ret)
			goto error;

		/* get existing config for present object */
		ret = json_object_object_get_ex(c, obj_name, &obj_config);
		if (ret == 0) {
			obj_config = json_object_new_object();
			json_object_object_add(c, obj_name, obj_config);
		}

		if (obj_config == NULL) {
			log_error("Unable to create new json object");
			ret = -ENOMEM;
			goto error;
		}

		/* merge another config with existing one */
		ret = assemble_config_from_dir(dir_path, obj_config);
		if (ret < 0) {
			log_error("Unable to assemble config for object %s", obj_name);
			json_object_put(obj_config);
			goto error;
		}
	}

	free(list);
	return 0;

error:
	while (num--)
		free(list[num]);
	free(list);
	return ret;
}

static int load_single_config(const char *config_path, struct json_object *c)
{
	int ret = 0;

	ret = access(config_path, F_OK | R_OK);
	if (ret) {
		/* file doesn't exist */
		if (errno == ENOENT) {
			log_debug("Config file doesn't exist %s", config_path);
			return -ENOENT;
		}

		/* file exists but we are unable to access it*/
		log_error("Could not access config file %s", config_path);
		return -errno;
	}

	/* file exists and we are able to access it*/
	log_debug("Using config file: %s", config_path);

	ret = json_object_merge_with_file(c, config_path);
	if (ret) {
		log_error("Unable to load config %s", config_path);
		return ret;
	}

	return 0;
}

int load_config(const char *src, int type, int optional, struct json_object *c)
{
	int ret;

	switch (type) {
	case CONFIG_TYPE_FILE:
		ret = load_single_config(src, c);
		break;

	case CONFIG_TYPE_DIR:
		ret = load_multidir_config(src, c);
		break;

	default:
		log_error("Wrong config type");
		return -EINVAL;
	}

	if ((ret < 0 && ret != -ENOENT) || (ret == -ENOENT && !optional))
		return ret;

	return 0;
}

int get_config_field(struct json_object *root, const char *key, void *value, int type)
{
	struct json_object *node = NULL;

	if (!json_object_object_get_ex(root, key, &node)) {
		log_debug("Config does not contain %s parameter. "
				  "Using default value", key);
		return 0;
	}

	if (!json_object_is_type(node, type)) {
		int t = json_object_get_type(node);

		log_error("Option %s has to be a %s, but is %s",
				  key, json_type_to_name(type), json_type_to_name(t));
		return -EINVAL;
	}

	switch (type) {
	case json_type_object: {
		*(struct json_object **) value = node;
		break;
	}

	case json_type_int: {
		int node_value;

		errno = 0;
		node_value = json_object_get_int(node);
		if (node_value == 0 && errno != 0) {
			log_error("Failed to get %s parameter from config file", key);
			return -EINVAL;
		}

		*(int *) value = node_value;
		break;
	}

	case json_type_string: {
		const char *node_value;
		char *str;

		node_value = json_object_get_string(node);
		str = strdup(node_value);
		if (str == NULL) {
			log_error("Failed to duplicate string");
			return -ENOMEM;
		}

		*(char **) value = str;
		break;
	}
	}

	return 0;
}
