/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <systemd/sd-bus.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <libsyscommon/dbus-system.h>

#include "common.h"
#include "systemd_dbus.h"
#include "log.h"

static sd_bus *systemd_bus = NULL;
static bool epc_systemd_private = false;

int epc_dbus_call_async(char *service, char *obj, char *interface, char *method,
					 epc_dbus_handler_t cb,
					 void *user_data, char *types, ...)
{
	dbus_handle_h bus = NULL;
	GVariant *var = NULL;

	bus = dbus_handle_get_connection(G_BUS_TYPE_SYSTEM, 0);
	if (!bus)
		return -EINVAL;


	if (types) {
		va_list args;

		va_start(args, types);
		var = g_variant_new_va(types, NULL, &args);
		va_end(args);
	}

	return dbus_handle_method_async_with_reply_var(service, obj, interface, method, var, cb, -1, user_data);
}

int epc_dbus_call(char *service, char *obj, char *interface, char *method,
					 char *types, ...)
{
	dbus_handle_h bus = NULL;
	GVariant *var = NULL;

	bus = dbus_handle_get_connection(G_BUS_TYPE_SYSTEM, 0);
	if (!bus)
		return -EINVAL;

	if (types) {
		va_list args;

		va_start(args, types);
		var = g_variant_new(types, NULL, args);
		va_end(args);
	}

	return dbus_handle_method_sync_var(service, obj, interface, method, var);
}

static int get_bus_by_address(const char *addr, sd_bus **bus)
{
	_cleanup_(sd_bus_unrefp) sd_bus *rbus = NULL;
	int ret;

	ret = sd_bus_new(&rbus);
	if (ret < 0)
		return ret;

	ret = sd_bus_set_address(rbus, addr);
	if (ret < 0)
		return ret;

	ret = sd_bus_start(rbus);
	if (ret < 0)
		return ret;

	*bus = rbus;
	rbus = NULL;
	return 0;
}

int epc_acquire_systemd_bus(sd_bus **bus)
{
	int ret;

	if (systemd_bus) {
		*bus = systemd_bus;
		sd_bus_ref(systemd_bus);
		return 0;
	}

	if (!epc_systemd_private) {
		ret = sd_bus_open_system(bus);
		if (ret < 0)
			return ret;
	} else {
		ret = get_bus_by_address(SYSTEMD_KERNEL_BUS, bus);
		if (ret < 0) {
			ret = get_bus_by_address(SYSTEMD_PRIVATE_BUS, bus);
			if (ret < 0)
				return ret;
		}
	}

	systemd_bus = *bus;

	return 0;
}

void epc_set_systemd_private(bool priv)
{
	epc_systemd_private = priv;
}
