/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <systemd/sd-bus.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timerfd.h>

#include "common.h"
#include "systemd_dbus.h"
#include "activationd_acceptance_tests.h"
#include "activationd_acceptance_tests_dbus_signal.h"
#include "activationd_acceptance_tests_unit_control.h"
#include "activationd_acceptance_tests_vconf.h"

#define DBUS_PROPERTIES_INTERFACE "org.freedesktop.DBus.Properties"
#define SERVICE_NAME_S_P "activationd-acceptance-test-s-p.service"
#define SERVICE_NAME_S_N "activationd-acceptance-test-s-n.service"
#define SERVICE_NAME_I_P "activationd-acceptance-test-i-p.service"
#define SERVICE_NAME_I_N "activationd-acceptance-test-i-n.service"
#define SERVICE_NAME_D_P "activationd-acceptance-test-d-p.service"
#define SERVICE_NAME_D_N "activationd-acceptance-test-d-n.service"
#define SERVICE_NAME_B_P "activationd-acceptance-test-b-p.service"
#define SERVICE_NAME_B_N "activationd-acceptance-test-b-n.service"
#define VCONF_KEY_S_P "file/activationd/acceptance_string_p"
#define VCONF_KEY_S_N "file/activationd/acceptance_string_n"
#define VCONF_KEY_I_P "file/activationd/acceptance_int_p"
#define VCONF_KEY_I_N "file/activationd/acceptance_int_n"
#define VCONF_KEY_D_P "file/activationd/acceptance_double_p"
#define VCONF_KEY_D_N "file/activationd/acceptance_double_n"
#define VCONF_KEY_B_P "file/activationd/acceptance_bool_p"
#define VCONF_KEY_B_N "file/activationd/acceptance_bool_n"

enum {
	TEST_TYPE_UNKNOWN,
	TEST_TYPE_VCONF,
	TEST_TYPE_DBUS_SIGNAL,
	TEST_TYPE_UNIT_CONTROL
};

static struct test_sequence unit_control_acceptance_tests[] = {
	{
		.desc = "unit control positive test",
		.actions = {
			{{
				{.type = VKC_STRING, .value.s = "Start"}
			 },
			{ACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_STRING, .value.s = "Restart"}
			 },
			{INACTIVE, ACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_STRING, .value.s = "Stop"}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestSP",
		.unit = SERVICE_NAME_S_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = unit_control_step
	},
	{
		.desc = "unit control negative test",
		.actions = {
			{{
				{.type = VKC_STRING, .value.s = "Start"}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_STRING, .value.s = "Restart"}
			 },
			{INACTIVE, INACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_STRING, .value.s = "Stoo"}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestIN",
		.unit = SERVICE_NAME_I_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = unit_control_step
	},
	{
		.desc = "unit control whitelist negative test",
		.actions = {
			{{
				{.type = VKC_STRING, .value.s = "Start"}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_STRING, .value.s = "Restart"}
			 },
			{INACTIVE, INACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_STRING, .value.s = "Stop"}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestSN",
		.unit = SERVICE_NAME_S_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = unit_control_step
	}
};

static struct test_sequence dbus_signal_acceptance_tests[] = {
	{
		.desc = "string (positive test)",
		.actions = {
			{{
				{.type = VKC_STRING, .value.s = "start"}
			 },
			{ACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_STRING, .value.s = "restart"}
			 },
			{INACTIVE, ACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_STRING, .value.s = "stop"}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestSP",
		.unit = SERVICE_NAME_S_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = dbus_signal_step
	},
	{
		.desc = "string (negative test)",
		.actions = {
			{{
				{.type = VKC_STRING, .value.s = "start"}
			 },
			{ACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_STRING, .value.s = "restart"}
			 },
			{INACTIVE, ACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_STRING, .value.s = "stop"}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestSN",
		.unit = SERVICE_NAME_S_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = dbus_signal_step
	},
	{
		.desc = "int (positive test)",
		.actions = {
			{{
				{.type = VKC_INT, .value.i = 1}
			 },
			{ACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_INT, .value.i = 2}
			 },
			{INACTIVE, ACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_INT, .value.i = 3}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestIP",
		.unit = SERVICE_NAME_I_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = dbus_signal_step
	},
	{
		.desc = "int (negative test)",
		.actions = {
			{{
				{.type = VKC_INT, .value.i = 1}
			 },
			{ACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_INT, .value.i = 2}
			 },
			{INACTIVE, ACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_INT, .value.i = 3}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestIN",
		.unit = SERVICE_NAME_I_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = dbus_signal_step
	},
	{
		.desc = "double (positive test)",
		.actions = {
			{{
				{.type = VKC_DOUBLE, .value.d = 1.0}
			 },
			{ACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_DOUBLE, .value.d = 2.0}
			 },
			{INACTIVE, ACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_DOUBLE, .value.d = 3.0}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestDP",
		.unit = SERVICE_NAME_D_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = dbus_signal_step
	},
	{
		.desc = "double (negative test)",
		.actions = {
			{{
				{.type = VKC_DOUBLE, .value.d = 1.0}
			 },
			{ACTIVE, UNKNOWN}, 1, 1, 0},
			{{
				{.type = VKC_DOUBLE, .value.d = 2.0}
			 },
			{INACTIVE, ACTIVE}, 1, 2, 0},
			{{
				{.type = VKC_DOUBLE, .value.d = 3.0}
			 },
			{INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = "AcceptanceTestDN",
		.unit = SERVICE_NAME_D_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = dbus_signal_step
	},
	{
		.desc = "bool (positive test)",
		.actions = {
			{{
				{.type = VKC_BOOL, .value.b = true},
				{.type = VKC_BOOL, .value.b = true}
			 },
			{ACTIVE, UNKNOWN}, 2, 1, 0},
			{{
				{.type = VKC_BOOL, .value.b = true},
				{.type = VKC_BOOL, .value.b = false}
			 },
			{INACTIVE, ACTIVE}, 2, 2, 0},
			{{
				{.type = VKC_BOOL, .value.b = false},
				{.type = VKC_BOOL, .value.b = false}
			 },
			{INACTIVE, UNKNOWN}, 2, 1, 0}
		},
		.id = "AcceptanceTestBP",
		.unit = SERVICE_NAME_B_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = dbus_signal_step
	},
	{
		.desc = "bool (negative test)",
		.actions = {
			{{
				{.type = VKC_BOOL, .value.b = true},
				{.type = VKC_BOOL, .value.b = true}
			 },
			{ACTIVE, UNKNOWN}, 2, 1, 0},
			{{
				{.type = VKC_BOOL, .value.b = true},
				{.type = VKC_BOOL, .value.b = false}
			 },
			{INACTIVE, ACTIVE}, 2, 2, 0},
			{{
				{.type = VKC_BOOL, .value.b = false},
				{.type = VKC_BOOL, .value.b = false}
			 },
			{INACTIVE, UNKNOWN}, 2, 1, 0}
		},
		.id = "AcceptanceTestBN",
		.unit = SERVICE_NAME_B_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = dbus_signal_step
	}
};

static struct test_sequence vconf_acceptance_tests[] = {
	{
		.desc = "string (positive test)",
		.actions = {
			{{{.type = VKC_STRING, .value.s = "START"}}, {ACTIVE, UNKNOWN}, 1, 1, 0},
			{{{.type = VKC_STRING, .value.s = "RELOAD"}}, {INACTIVE, ACTIVE}, 1, 2, 0},
			{{{.type = VKC_STRING, .value.s = "STOP"}}, {INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = VCONF_KEY_S_P,
		.unit = SERVICE_NAME_S_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = vconf_step
	},
	{
		.desc = "string (negative test)",
		.actions = {
			{{{.type = VKC_STRING, .value.s = "START"}}, {ACTIVE, UNKNOWN}, 1, 1, 0},
			{{{.type = VKC_STRING, .value.s = "RELOAD"}}, {INACTIVE, ACTIVE}, 1, 2, 0},
			{{{.type = VKC_STRING, .value.s = "STOP"}}, {INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = VCONF_KEY_S_N,
		.unit = SERVICE_NAME_S_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = vconf_step
	},
	{
		.desc = "int (positive test)",
		.actions = {
			{{{.type = VKC_INT, .value.i = 1}}, {ACTIVE, UNKNOWN}, 1, 1, 0},
			{{{.type = VKC_INT, .value.i = 2}}, {INACTIVE, ACTIVE}, 1, 2, 0},
			{{{.type = VKC_INT, .value.i = 3}}, {INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = VCONF_KEY_I_P,
		.unit = SERVICE_NAME_I_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = vconf_step
	},
	{
		.desc = "int (negative test)",
		.actions = {
			{{{.type = VKC_INT, .value.i = 1}}, {ACTIVE, UNKNOWN}, 1, 1, 0},
			{{{.type = VKC_INT, .value.i = 2}}, {INACTIVE, ACTIVE}, 1, 2, 0},
			{{{.type = VKC_INT, .value.i = 3}}, {INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = VCONF_KEY_I_N,
		.unit = SERVICE_NAME_I_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = vconf_step
	},
	{
		.desc = "double (positive test)",
		.actions = {
			{{{.type = VKC_DOUBLE, .value.d = 1.1}}, {ACTIVE, UNKNOWN}, 1, 1, 0},
			{{{.type = VKC_DOUBLE, .value.d = 1.2}}, {INACTIVE, ACTIVE}, 1, 2, 0},
			{{{.type = VKC_DOUBLE, .value.d = 1.3}}, {INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = VCONF_KEY_D_P,
		.unit = SERVICE_NAME_D_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = vconf_step
	},
	{
		.desc = "double (negative test)",
		.actions = {
			{{{.type = VKC_DOUBLE, .value.d = 1.1}}, {ACTIVE, UNKNOWN}, 1, 1, 0},
			{{{.type = VKC_DOUBLE, .value.d = 1.2}}, {INACTIVE, ACTIVE}, 1, 2, 0},
			{{{.type = VKC_DOUBLE, .value.d = 1.3}}, {INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = VCONF_KEY_D_N,
		.unit = SERVICE_NAME_D_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = vconf_step
	},
	{
		.desc = "bool (positive test)",
		.actions = {
			{{{.type = VKC_BOOL, .value.b = true}}, {ACTIVE, UNKNOWN}, 1, 1, 0},
			{{{.type = VKC_BOOL, .value.b = true}}, {INACTIVE, ACTIVE}, 1, 2, 0},
			{{{.type = VKC_BOOL, .value.b = false}}, {INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = VCONF_KEY_B_P,
		.unit = SERVICE_NAME_B_P,
		.path = NULL,
		.expected_result = TEST_SUCCESS,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = vconf_step
	},
	{
		.desc = "bool (negative test)",
		.actions = {
			{{{.type = VKC_BOOL, .value.b = true}}, {ACTIVE, UNKNOWN}, 1, 1, 0},
			{{{.type = VKC_BOOL, .value.b = true}}, {INACTIVE, ACTIVE}, 1, 2, 0},
			{{{.type = VKC_BOOL, .value.b = false}}, {INACTIVE, UNKNOWN}, 1, 1, 0}
		},
		.id = VCONF_KEY_B_N,
		.unit = SERVICE_NAME_B_N,
		.path = NULL,
		.expected_result = TEST_FAIL,
		.result = TEST_UNKNOWN,
		.current_action = 0,
		.step = vconf_step
	}
};

static const char *test_result(int r)
{
	if (r == TEST_SUCCESS)
		return "success";
	else if (r == TEST_FAIL)
		return "fail";
	else
		return "unknown";
}

static void cleanup(struct test_data *data)
{
	for (int i = 0; i < data->tests_count; i++)
		free(data->tests[i].path);
	if (!data->bus)
		return;
	sd_bus_flush(data->bus);
	sd_bus_unref(data->bus);
}

static int signal_handler(sd_event_source *s,
		const struct signalfd_siginfo *si,
		void *userdata)
{
	sd_event *loop = userdata;
	sd_event_exit(loop, EXIT_FAILURE);
	return 0;
}

static int start(sd_event_source *s, uint64_t usec, void *userdata)
{
	struct test_data *data = (struct test_data*)userdata;
	return start_test(data);
}

static int timeout_handler(sd_event_source *s, uint64_t usec, void *userdata)
{
	struct test_data *data = (struct test_data*)userdata;
	fprintf(stderr, "test timed out\n");
	sd_event_exit(data->loop, EXIT_FAILURE);
	return 0;
}

static int call_subscribe(struct test_data *data)
{
	__attribute__((cleanup(sd_bus_message_unrefp))) sd_bus_message *reply = NULL;
	__attribute__((cleanup(sd_bus_error_free))) sd_bus_error error = SD_BUS_ERROR_NULL;
	int ret;

	ret = sd_bus_call_method(data->bus,
			SYSTEMD_SERVICE,
			SYSTEMD_OBJ,
			SYSTEMD_MANAGER_INTERFACE,
			"Subscribe",
			&error,
			&reply,
			"");

	if (ret < 0) {
		fprintf(stderr, "Failed to issue Subscribe() call: %s\n", error.message);
		return -1;
	}
	return 0;
}

static int call_load_unit(struct test_data *data)
{
	__attribute__((cleanup(sd_bus_message_unrefp))) sd_bus_message *reply = NULL;
	__attribute__((cleanup(sd_bus_error_free))) sd_bus_error error = SD_BUS_ERROR_NULL;
	int ret;
	const char *path = NULL;
	char err[512];

	for (int i = 0; i < data->tests_count; i++) {
		ret = sd_bus_call_method(data->bus,
				SYSTEMD_SERVICE,
				SYSTEMD_OBJ,
				SYSTEMD_MANAGER_INTERFACE,
				"LoadUnit",
				&error,
				&reply,
				"s",
				data->tests[i].unit);

		if (ret < 0) {
			fprintf(stderr, "Failed to issue LoadUnit() call: %s\n", error.message);
			return -1;
		}

		ret = sd_bus_message_read(reply, "o", &path);
		if (ret < 0) {
			if (strerror_r(-ret, err, sizeof err))
				fprintf(stderr, "Failed to read reply: %s\n", err);
			return -1;
		}

		if (!path) {
			fprintf(stderr, "could not get object path\n");
			return -1;
		}
		data->tests[i].path = strdup(path);
	}
	return 0;
}

static int print_results(struct test_data *data)
{
	int ret = 0;
	for (int i = 0; i < data->tests_count; i++) {
		printf("testing %s - expected %s, got %s\n",
				data->tests[i].desc,
				test_result(data->tests[i].expected_result),
				test_result(data->tests[i].result));
		if (data->tests[i].expected_result != data->tests[i].result)
			ret = 1;
	}
	return ret;
}

static void print_usage(const char *name)
{
	printf("usage: %s -t vconf|signal|unit_control\n", name);
	return;
}

static int parse_cmdline(int argc, char *argv[], int *type)
{
	int option;

	*type = TEST_TYPE_UNKNOWN;

	while ((option = getopt(argc, argv, "t:")) != -1) {
		switch (option) {
			case 't':
				if (!strcmp(optarg, "vconf"))
					*type = TEST_TYPE_VCONF;
				else if (!strcmp(optarg, "signal"))
					*type = TEST_TYPE_DBUS_SIGNAL;
				else if (!strcmp(optarg, "unit_control"))
					*type = TEST_TYPE_UNIT_CONTROL;
				break;
			default:
				return -EINVAL;
		}
	}

	if (*type == TEST_TYPE_UNKNOWN)
		return -EINVAL;

	return 0;
}


int main(int argc, char *argv[])
{
	struct test_data data = {
		.bus = NULL,
		.loop = NULL,
		.step_timer = NULL,
		.state = UNKNOWN,
		.prev_state = UNKNOWN,
		.current_test = 0
	};

	int test_type;
	int r;
	uint64_t now;
	sigset_t ss;
	char buf[512];
	char err[512];

	r = parse_cmdline(argc, argv, &test_type);
	if (r < 0) {
		print_usage(argv[0]);
		goto fail;
	}

	if (test_type == TEST_TYPE_VCONF) {
		data.tests = vconf_acceptance_tests;
		data.tests_count = ARRAY_SIZE(vconf_acceptance_tests);
	} else if (test_type == TEST_TYPE_DBUS_SIGNAL) {
		data.tests = dbus_signal_acceptance_tests;
		data.tests_count = ARRAY_SIZE(dbus_signal_acceptance_tests);
	} else if (test_type == TEST_TYPE_UNIT_CONTROL) {
		data.tests = unit_control_acceptance_tests;
		data.tests_count = ARRAY_SIZE(unit_control_acceptance_tests);
	} else {
		fprintf(stderr, "Unknown test type\n");
		goto fail;
	}

	r = sd_bus_open_system(&data.bus);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to read reply: %s\n", err);
		goto fail;
	}

	r = call_subscribe(&data);
	if (r < 0)
		goto fail;

	r = call_load_unit(&data);
	if (r < 0)
		goto fail;

	for (int i = 0; i < data.tests_count; i++) {
		r = snprintf(buf, sizeof buf, "type='signal',interface='org.freedesktop.DBus.Properties',member='PropertiesChanged', path='%s'", data.tests[i].path);
		if (r < 0) {
			if (strerror_r(-r, err, sizeof err))
				fprintf(stderr, "Failed to create match rule: %s\n", err);
			goto fail;
		}

		r = sd_bus_add_match(data.bus, NULL, buf, on_properties_changed, &data);
		if (r < 0) {
			if (strerror_r(-r, err, sizeof err))
				fprintf(stderr, "Failed to add PropertiesChanged match: %s\n", err);
			goto fail;
		}
	}

	r = sd_event_new(&data.loop);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to create main loop: %s\n", err);
		goto fail;
	}

	sigemptyset(&ss);
	sigaddset(&ss, SIGINT);
	sigaddset(&ss, SIGTERM);
	r = sigprocmask(SIG_BLOCK, &ss, NULL);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to change signal mask: %s\n", err);
		goto fail;
	}

	r = sd_event_add_signal(data.loop, NULL, SIGINT, signal_handler, data.loop);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to register SIGINT handler: %s\n", err);
		goto fail;
	}

	r = sd_event_add_signal(data.loop, NULL, SIGTERM, signal_handler, data.loop);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to register SIGTERM handler: %s\n", err);
		goto fail;
	}

	r = sd_bus_attach_event(data.bus, data.loop, 0);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to attach bus to event loop: %s\n", err);
		goto fail;
	}

	r = sd_event_add_time(data.loop, NULL, CLOCK_REALTIME, 0, 0, start, &data);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to attach timer to event loop: %s\n", err);
		goto fail;
	}

	/* time out all tests after 30sec */
	sd_event_now(data.loop, CLOCK_REALTIME, &now);

	r = sd_event_add_time(data.loop, NULL, CLOCK_REALTIME, now+30000000, 0, timeout_handler, &data);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to attach timer to event loop: %s\n", err);
		goto fail;
	}

	r = sd_event_loop(data.loop);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to run main loop: %s\n", err);
		goto fail;
	}

	r = print_results(&data);
	cleanup(&data);
	return r;

fail:
	cleanup(&data);
	return EXIT_FAILURE;
}
