/*
 * This file is part of activationd.
 *
 * Copyright © 2019 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "vconf_key_changed_event.h"
#include "activationd_acceptance_tests.h"
#include "activationd_acceptance_tests_dbus_signal.h"
#include "activationd_acceptance_tests_vconf.h"

int test_step(struct test_data *data);

static const char *from_unit_state(enum unit_state state)
{
	switch (state) {
	case UNKNOWN:
		return "unknown";
	case ACTIVE:
		return "active";
	default:
		return "inactive";
	}
}

static enum unit_state to_unit_state(const char *state)
{
	if (!strncmp (state, "active", 6))
		return ACTIVE;
	else if (!strncmp (state, "inactive", 8) || !strncmp (state, "deactivating", 12))
		return INACTIVE;
	else
		return UNKNOWN;
}

static void next_test(struct test_data *data)
{
	data->state = UNKNOWN;
	data->prev_state = UNKNOWN;
	data->current_test++;

	test_step(data);
}

static void stop_test_step_timeout(struct test_data *data)
{
	int enabled;
	sd_event_source_get_enabled(data->step_timer, &enabled);

	if (enabled != SD_EVENT_OFF)
		sd_event_source_set_enabled(data->step_timer, SD_EVENT_OFF);
}

static int on_test_step_timeout(sd_event_source *s, uint64_t usec, void *userdata)
{
	struct test_data *data = (struct test_data *)userdata;
	struct test_sequence *test = &data->tests[data->current_test];

	test->result = TEST_FAIL;
	next_test(data);
	return 0;
}

static int start_test_step_timeout(struct test_data *data)
{
	int r;
	uint64_t now;
	char err[512];

	sd_event_now(data->loop, CLOCK_REALTIME, &now);
	r = sd_event_add_time(data->loop, &data->step_timer, CLOCK_REALTIME, now+2000000, 0, on_test_step_timeout, data);
	if (r < 0) {
		if (strerror_r(-r, err, sizeof err))
			fprintf(stderr, "Failed to attach timer to event loop: %s\n", err);
		return r;
	}

	return 0;
}

static int check_test_result(struct test_data *data)
{
	struct test_sequence *test = &data->tests[data->current_test];
	struct action *a = &test->actions[test->current_action];
	enum unit_state exp_state = a->expected_states[a->current_state];


	if (data->state != exp_state) {
		fprintf(stderr, "test failed (%s instead of %s at step %d of action %d of test %d)\n",
				from_unit_state(data->state),
				from_unit_state(exp_state),
				a->current_state,
				test->current_action,
				data->current_test);
		test->result = TEST_FAIL;
		next_test(data);
		return -1;
	}
	a->current_state++;
	if (a->current_state == a->state_count) {
		test->current_action++;
		if (test_step(data) < 0)
			return -1;
	}
	return 0;
}

static int check_unit_state(struct test_data *data, const char *state)
{
	enum unit_state new_state;

	new_state = to_unit_state(state);
	if (data->state != new_state) {
		data->prev_state = data->state;
		data->state = new_state;
		check_test_result(data);
	}
	return 0;
}

int test_step(struct test_data *data)
{
	struct test_sequence *test;

	stop_test_step_timeout(data);

	/* last test */
	if (data->current_test == data->tests_count) {
		sd_event_exit(data->loop, EXIT_SUCCESS);
		return 0;
	}
	test = &data->tests[data->current_test];

	if (test->current_action == ARRAY_SIZE(test->actions)) {
		test->result = TEST_SUCCESS;
		next_test(data);
		return 0;
	}

	//if (set_vconf_key(test->key, test->actions[test->current_action].action))
	//	return -1;
	if (!test->step || test->step(data) != 0)
		return -1;

	start_test_step_timeout(data);
	return 0;
}

int start_test(struct test_data *data)
{
	return test_step(data);
}


int on_properties_changed(sd_bus_message *m, void *userdata, sd_bus_error *ret_error)
{
	int r;
	struct test_data *data = (struct test_data*)userdata;
	const char *iface;
	const char *member;
	const char *value = NULL;
	const char *contents;

	r = sd_bus_message_read(m, "s", &iface);
	if (r < 0) {
		fprintf(stderr, "Failed to parse PropertiesChanged\n");
		return r;
	}

	r = sd_bus_message_enter_container(m, SD_BUS_TYPE_ARRAY, "{sv}");
	if (r < 0)
		return r;

	while ((r = sd_bus_message_enter_container(m, SD_BUS_TYPE_DICT_ENTRY, "sv")) > 0) {
		char type;

		r = sd_bus_message_read(m, "s", &member);
		if (r < 0)
			return r;

		r = sd_bus_message_peek_type(m, &type, &contents);
		if (r < 0)
			return r;

		r = sd_bus_message_enter_container(m, SD_BUS_TYPE_VARIANT, contents);
		if (r < 0)
			return r;

		if (strcmp(member, "ActiveState") == 0) {
			r = sd_bus_message_read(m, "s", &value);
			if (r < 0)
				return r;
		} else {
			r = sd_bus_message_skip(m, contents);
			if (r < 0)
				return r;
		}

		r = sd_bus_message_exit_container(m);
		if (r < 0)
			return r;

		r = sd_bus_message_exit_container(m);
		if (r < 0)
			return r;
	}

	r = sd_bus_message_exit_container(m);
	if (r < 0)
		return r;

	if (value)
		check_unit_state(data, value);
	return 0;
}

